@@ -37,21 +37,21 @@
 int gli_utf8input = FALSE ;
 int gli_utf8output = FALSE ;
 
-char *gli_conf_propr = "CharterBT-Roman";
-char *gli_conf_propb = "CharterBT-Bold";
-char *gli_conf_propi = "CharterBT-Italic";
-char *gli_conf_propz = "CharterBT-BoldItalic";
+char *gli_conf_propr = "NotoSerif-Regular";
+char *gli_conf_propb = "NotoSerif-Bold";
+char *gli_conf_propi = "NotoSerif-Italic";
+char *gli_conf_propz = "NotoSerif-BoldItalic";
 
-char *gli_conf_monor = "LuxiMonoRegular";
-char *gli_conf_monob = "LuxiMonoBold";
-char *gli_conf_monoi = "LuxiMonoOblique";
-char *gli_conf_monoz = "LuxiMonoBoldOblique";
+char *gli_conf_monor = "GoMono-Regular";
+char *gli_conf_monob = "GoMono-Bold";
+char *gli_conf_monoi = "GoMono-Italic";
+char *gli_conf_monoz = "GoMono-BoldItalic";
 
 #ifdef BUNDLED_FONTS
 char *gli_conf_monofont = "";
 char *gli_conf_propfont = "";
-float gli_conf_monosize = 12.6;	/* good size for LuxiMono */
-float gli_conf_propsize = 14.7;	/* good size for CharterBT */
+float gli_conf_monosize = 12.5;	/* good size for GoMono */
+float gli_conf_propsize = 13.4;	/* good size for NotoSerif */
 #else
 char *gli_conf_monofont = "Liberation Mono";
 char *gli_conf_propfont = "Linux Libertine O";
@@ -147,9 +147,15 @@ int gli_wbordery = 1;
 int gli_tmarginx = 7;
 int gli_tmarginy = 7;
 
+int gli_wmarginx_save = 15;
+int gli_wmarginy_save = 15;
+
 int gli_cols = 60;
 int gli_rows = 25;
 
+int gli_conf_lockcols = FALSE;
+int gli_conf_lockrows = FALSE;
+
 float gli_conf_propaspect = 1.0;
 float gli_conf_monoaspect = 1.0;
 
@@ -158,14 +164,19 @@ int gli_leading = 20;
 
 int gli_conf_justify = 0;
 int gli_conf_quotes = 1;
+int gli_conf_dashes = 1;
 int gli_conf_spaces = 0;
 int gli_conf_caps = 0;
 
 int gli_conf_graphics = 1;
 int gli_conf_sound = 1;
 int gli_conf_speak = 0;
+int gli_conf_speak_input = 0;
+const char *gli_conf_speak_language = NULL;
 
-int gli_conf_stylehint = 0;
+int gli_conf_fullscreen = 0;
+
+int gli_conf_stylehint = 1;
 int gli_conf_safeclicks = 0;
 
 static void parsecolor(char *str, unsigned char *rgb)
@@ -186,7 +197,7 @@ static void parsecolor(char *str, unsigned char *rgb)
     rgb[2] = strtol(b, NULL, 16);
 }
 
-char* trim(char* src)
+static char* trim(char* src)
 {
     while(src[strlen(src)-1] == ' ' || src[strlen(src)-1] == '\t')
         src[strlen(src)-1] = 0;
@@ -311,10 +322,53 @@ static void readoneconfig(char *fname, char *argv0, char *gamefile)
         if (!strcmp(cmd, "cols"))
             gli_cols = atoi(arg);
 
+        if (!strcmp(cmd, "minrows"))
+        {
+            int r = atoi(arg);
+            if (gli_rows < r)
+                gli_rows = r;
+        }
+
+        if (!strcmp(cmd, "maxrows"))
+        {
+            int r = atoi(arg);
+            if (gli_rows > r)
+                gli_rows = r;
+        }
+
+        if (!strcmp(cmd, "mincols"))
+        {
+            int c = atoi(arg);
+            if (gli_cols < c)
+                gli_cols = c;
+        }
+
+        if (!strcmp(cmd, "maxcols"))
+        {
+            int c = atoi(arg);
+            if (gli_cols > c)
+                gli_cols = c;
+        }
+
+        if (!strcmp(cmd, "lockrows"))
+            gli_conf_lockrows = atoi(arg);
+        if (!strcmp(cmd, "lockcols"))
+            gli_conf_lockcols = atoi(arg);
+
+        if (!strcmp(cmd, "wmarginx"))
+        {
+            gli_wmarginx = atoi(arg);
+            gli_wmarginx_save = gli_wmarginx;
+        }
+
+        if (!strcmp(cmd, "wmarginy"))
+        {
+            gli_wmarginy = atoi(arg);
+            gli_wmarginy_save = gli_wmarginy;
+        }
+
         if (!strcmp(cmd, "wpaddingx")) gli_wpaddingx = atoi(arg);
         if (!strcmp(cmd, "wpaddingy")) gli_wpaddingy = atoi(arg);
-        if (!strcmp(cmd, "wmarginx")) gli_wmarginx = atoi(arg);
-        if (!strcmp(cmd, "wmarginy")) gli_wmarginy = atoi(arg);
         if (!strcmp(cmd, "wborderx")) gli_wborderx = atoi(arg);
         if (!strcmp(cmd, "wbordery")) gli_wbordery = atoi(arg);
         if (!strcmp(cmd, "tmarginx")) gli_tmarginx = atoi(arg);
@@ -367,6 +421,8 @@ static void readoneconfig(char *fname, char *argv0, char *gamefile)
             gli_conf_justify = atoi(arg);
         if (!strcmp(cmd, "quotes"))
             gli_conf_quotes = atoi(arg);
+        if (!strcmp(cmd, "dashes"))
+            gli_conf_dashes = atoi(arg);
         if (!strcmp(cmd, "spaces"))
             gli_conf_spaces = atoi(arg);
         if (!strcmp(cmd, "caps"))
@@ -376,8 +432,16 @@ static void readoneconfig(char *fname, char *argv0, char *gamefile)
             gli_conf_graphics = atoi(arg);
         if (!strcmp(cmd, "sound"))
             gli_conf_sound = atoi(arg);
+
+        if (!strcmp(cmd, "fullscreen"))
+            gli_conf_fullscreen = atoi(arg);
+
         if (!strcmp(cmd, "speak"))
             gli_conf_speak = atoi(arg);
+        if (!strcmp(cmd, "speak_input"))
+            gli_conf_speak_input = atoi(arg);
+        if (!strcmp(cmd, "speak_language"))
+            gli_conf_speak_language = strdup(arg);
 
         if (!strcmp(cmd, "stylehint"))
             gli_conf_stylehint = atoi(arg);
@@ -430,7 +494,6 @@ void gli_read_config(int argc, char **argv)
     char argv0[1024] = "default";
     char buf[1024];
     int i;
-    char *tmp;
 
     /* load argv0 with name of executable without suffix */
     if (strrchr(argv[0], '\\'))
@@ -499,13 +562,6 @@ void gli_read_config(int argc, char **argv)
         readoneconfig(buf, argv0, gamefile);
     }
 
-    tmp = (char *) getcwd(buf, sizeof buf);
-    if (tmp)
-    {
-        strcat(buf, "/garglk.ini");
-        readoneconfig(buf, argv0, gamefile);
-    }
-
     if (argc > 1)
     {
         strcpy(buf, argv[argc-1]);
@@ -551,11 +607,9 @@ void gli_startup(int argc, char *argv[])
     if (!gli_baseline)
         gli_baseline = gli_conf_propsize + 0.5;
 
-#ifdef USETTS
     gli_initialize_tts();
     if (gli_conf_speak)
         gli_conf_quotes = 0;
-#endif
 
     gli_initialize_misc();
     gli_initialize_fonts();
