@@ -28,6 +28,8 @@
  * http://www.eblong.com/zarf/glk/index.html
  */
 
+#include <stddef.h>
+
 #include "gi_dispa.h"
 
 /* First, we define our own TRUE and FALSE and NULL, because ANSI
@@ -211,6 +213,8 @@ extern int gli_wbordery;
 
 extern int gli_wmarginx;
 extern int gli_wmarginy;
+extern int gli_wmarginx_save;
+extern int gli_wmarginy_save;
 extern int gli_wpaddingx;
 extern int gli_wpaddingy;
 extern int gli_tmarginx;
@@ -220,18 +224,27 @@ extern int gli_conf_lcd;
 
 extern int gli_conf_graphics;
 extern int gli_conf_sound;
+
+extern int gli_conf_fullscreen;
+
 extern int gli_conf_speak;
+extern int gli_conf_speak_input;
+extern const char *gli_conf_speak_language;
 
 extern int gli_conf_stylehint;
 extern int gli_conf_safeclicks;
 
 extern int gli_conf_justify;
 extern int gli_conf_quotes;
+extern int gli_conf_dashes;
 extern int gli_conf_spaces;
 extern int gli_conf_caps;
 
-extern int gli_rows;
 extern int gli_cols;
+extern int gli_rows;
+
+extern int gli_conf_lockcols;
+extern int gli_conf_lockrows;
 
 extern unsigned char gli_scroll_bg[3];
 extern unsigned char gli_scroll_fg[3];
@@ -446,8 +459,9 @@ struct window_textgrid_s
 
     /* for line input */
     void *inbuf;	/* unsigned char* for latin1, glui32* for unicode */
+    int inunicode;
     int inorgx, inorgy;
-    int inmax;
+    int inoriglen, inmax;
     int incurs, inlen;
     attr_t origattr;
     gidispatch_rock_t inarrayrock;
@@ -500,6 +514,7 @@ struct window_textbuffer_s
 
     /* for line input */
     void *inbuf;	/* unsigned char* for latin1, glui32* for unicode */
+    int inunicode;
     int inmax;
     long infence;
     long incurs;
@@ -556,7 +571,9 @@ struct glk_schannel_struct
 
 extern void gli_initialize_sound(void);
 extern void gli_initialize_tts(void);
-extern void gli_speak_tts(char *buf, int len, int interrupt);
+extern void gli_tts_speak(const glui32 *buf, size_t len);
+extern void gli_tts_flush(void);
+extern void gli_tts_purge(void);
 
 /* ---------------------------------------------------------------------- */
 /*
@@ -602,7 +619,7 @@ extern void win_textbuffer_cancel_line(window_t *win, event_t *ev);
 extern void win_textbuffer_click(window_textbuffer_t *dwin, int x, int y);
 extern void gcmd_buffer_accept_readchar(window_t *win, glui32 arg);
 extern void gcmd_buffer_accept_readline(window_t *win, glui32 arg);
-extern void gcmd_accept_scroll(window_t *win, glui32 arg);
+extern int gcmd_accept_scroll(window_t *win, glui32 arg);
 
 /* Declarations of library internal functions. */
 
@@ -619,9 +636,11 @@ extern void gli_window_redraw(window_t *win);
 extern void gli_window_put_char_uni(window_t *win, glui32 ch);
 extern int gli_window_unput_char_uni(window_t *win, glui32 ch);
 extern int gli_window_check_terminator(glui32 ch);
+extern void gli_window_refocus(window_t *win);
 
 extern void gli_windows_redraw(void);
 extern void gli_windows_size_change(void);
+extern void gli_windows_unechostream(stream_t *str);
 
 extern void gli_window_click(window_t *win, int x, int y);
 
@@ -633,6 +652,10 @@ void gli_input_next_focus();
 void gli_input_scroll_focus();
 void gli_input_handle_key(glui32 key);
 void gli_input_handle_click(int x, int y);
+void gli_input_guess_focus(void);
+void gli_input_more_focus(void);
+void gli_input_next_focus(void);
+void gli_input_scroll_focus(void);
 void gli_event_store(glui32 type, window_t *win, glui32 val1, glui32 val2);
 
 extern stream_t *gli_new_stream(glui32 type, int readable, int writable,
@@ -646,13 +669,19 @@ extern void gli_stream_fill_result(stream_t *str,
     stream_result_t *result);
 extern void gli_stream_echo_line(stream_t *str, char *buf, glui32 len);
 extern void gli_stream_echo_line_uni(stream_t *str, glui32 *buf, glui32 len);
+extern void gli_streams_close_all(void);
 
 extern fileref_t *gli_new_fileref(char *filename, glui32 usage,
     glui32 rock);
 extern void gli_delete_fileref(fileref_t *fref);
 
+#ifdef BUNDLED_FONTS
+void gli_get_builtin_font(int idx, const unsigned char **ptr, unsigned int *len);
+#endif
+
 void gli_initialize_fonts(void);
 void gli_draw_pixel(int x, int y, unsigned char alpha, unsigned char *rgb);
+void gli_draw_pixel_lcd(int x, int y, unsigned char *alpha, unsigned char *rgb);
 void gli_draw_clear(unsigned char *rgb);
 void gli_draw_rect(int x, int y, int w, int h, unsigned char *rgb);
 int gli_draw_string(int x, int y, int f, unsigned char *rgb, unsigned char *text, int len, int spacewidth);
@@ -694,8 +723,14 @@ picture_t *gli_picture_load(unsigned long id);
 void gli_picture_store(picture_t *pic);
 picture_t *gli_picture_retrieve(unsigned long id, int scaled);
 picture_t *gli_picture_scale(picture_t *src, int destwidth, int destheight);
+void gli_picture_increment(picture_t *pic);
+void gli_picture_decrement(picture_t *pic);
+piclist_t *gli_piclist_search(unsigned long id);
+void gli_piclist_clear(void);
 void gli_piclist_increment(void);
 void gli_piclist_decrement(void);
+void gli_picture_store_original(picture_t *pic);
+void gli_picture_store_scaled(picture_t *pic);
 
 window_graphics_t *win_graphics_create(window_t *win);
 void win_graphics_destroy(window_graphics_t *cutwin);
