@@ -27,6 +27,8 @@
 #import <Cocoa/Cocoa.h>
 #import <OpenGL/gl.h>
 #import <OpenGL/glu.h>
+#include <mach-o/dyld.h>
+#import "sysmac.h"
 
 #ifdef __ppc__
 #define ByteOrderOGL GL_UNSIGNED_INT_8_8_8_8
@@ -145,6 +147,8 @@ char *winfilters[] =
 - (IBAction) cut: (id) sender;
 - (IBAction) copy: (id) sender;
 - (IBAction) paste: (id) sender;
+- (void) moveWordBackward: (id) sender;
+- (void) moveWordForward: (id) sender;
 - (IBAction) performZoom: (id) sender;
 - (void) performRefresh: (NSNotification *) notice;
 - (NSString *) openFileDialog: (NSString *) prompt
@@ -196,6 +200,101 @@ char *winfilters[] =
     return self;
 }
 
+static BOOL isTextbufferEvent(NSEvent * evt)
+{
+    if ([evt type] != NSKeyDown)
+    {
+        return NO;
+    }
+
+    /*
+     * The following mirrors the checks in sysmac.m:winkey().
+     */
+
+    /* check for arrow keys */
+    if ([evt modifierFlags] & NSFunctionKeyMask)
+    {
+        /* alt/option modified key */
+        if ([evt modifierFlags] & NSAlternateKeyMask)
+        {
+            switch ([evt keyCode])
+            {
+                case NSKEY_LEFT  : return NO;
+                case NSKEY_RIGHT : return NO;
+                case NSKEY_DOWN  : return NO;
+                case NSKEY_UP    : return NO;
+                default: break;
+            }
+        }
+
+        /* command modified key */
+        if ([evt modifierFlags] & NSCommandKeyMask)
+        {
+            switch ([evt keyCode])
+            {
+                case NSKEY_LEFT  : return NO;
+                case NSKEY_RIGHT : return NO;
+                case NSKEY_DOWN  : return NO;
+                case NSKEY_UP    : return NO;
+                default: break;
+            }
+        }
+
+        /* unmodified key for line editing */
+        switch ([evt keyCode])
+        {
+            case NSKEY_LEFT  : return NO;
+            case NSKEY_RIGHT : return NO;
+            case NSKEY_DOWN  : return NO;
+            case NSKEY_UP    : return NO;
+            default: break;
+        }
+    }
+
+    /* check for menu commands */
+    if ([evt modifierFlags] & NSCommandKeyMask)
+    {
+        switch ([evt keyCode])
+        {
+            case NSKEY_X:
+            case NSKEY_C:
+                return NO;
+
+            case NSKEY_V:
+                return NO;
+
+            default: break;
+        }
+    }
+
+    /* check for command keys */
+    switch ([evt keyCode])
+    {
+        case NSKEY_PGUP : return NO;
+        case NSKEY_PGDN : return NO;
+        case NSKEY_HOME : return NO;
+        case NSKEY_END  : return NO;
+        case NSKEY_DEL  : return NO;
+        case NSKEY_BACK : return NO;
+        case NSKEY_ESC  : return NO;
+        case NSKEY_F1   : return NO;
+        case NSKEY_F2   : return NO;
+        case NSKEY_F3   : return NO;
+        case NSKEY_F4   : return NO;
+        case NSKEY_F5   : return NO;
+        case NSKEY_F6   : return NO;
+        case NSKEY_F7   : return NO;
+        case NSKEY_F8   : return NO;
+        case NSKEY_F9   : return NO;
+        case NSKEY_F10  : return NO;
+        case NSKEY_F11  : return NO;
+        case NSKEY_F12  : return NO;
+        default: break;
+    }
+
+    return YES;
+}
+
 - (void) sendEvent: (NSEvent *) event
 {
     int store = NO;
@@ -243,6 +342,10 @@ char *winfilters[] =
     if (store)
     {
         [eventlog insertObject: [event copy] atIndex:0];
+        if (isTextbufferEvent(event))
+        {
+            [textbuffer interpretKeyEvents: [NSArray arrayWithObject: event]];
+        }
         kill(processID, SIGUSR1);
     }
 
@@ -269,7 +372,8 @@ char *winfilters[] =
 
 - (void) sendChars: (NSEvent *) event
 {
-    [textbuffer interpretKeyEvents: [NSArray arrayWithObject:event]];
+    /* Does not work on macOS 10.12 (Sierra) */
+    //[textbuffer interpretKeyEvents: [NSArray arrayWithObject:event]];
 }
 
 - (NSString *) retrieveChars
@@ -293,7 +397,7 @@ char *winfilters[] =
                                     characters: @"X"
                    charactersIgnoringModifiers: @"X"
                                      isARepeat: NO
-                                       keyCode: 0x07]];
+                                       keyCode: NSKEY_X]];
 }
 
 - (IBAction) copy: (id) sender
@@ -307,7 +411,7 @@ char *winfilters[] =
                                     characters: @"C"
                    charactersIgnoringModifiers: @"C"
                                      isARepeat: NO
-                                       keyCode: 0x08]];
+                                       keyCode: NSKEY_C]];
 }
 
 - (IBAction) paste: (id) sender
@@ -321,7 +425,35 @@ char *winfilters[] =
                                     characters: @"V"
                    charactersIgnoringModifiers: @"V"
                                      isARepeat: NO
-                                       keyCode: 0x09]];
+                                       keyCode: NSKEY_V]];
+}
+
+- (void) moveWordBackward: (id) sender
+{
+    [self sendEvent: [NSEvent keyEventWithType: NSKeyDown
+                                      location: NSZeroPoint
+                                 modifierFlags: NSAlternateKeyMask
+                                     timestamp: NSTimeIntervalSince1970
+                                  windowNumber: [self windowNumber]
+                                       context: [self graphicsContext]
+                                    characters: @""
+                   charactersIgnoringModifiers: @""
+                                     isARepeat: NO
+                                       keyCode: NSKEY_LEFT]];
+}
+
+- (void) moveWordForward: (id) sender
+{
+    [self sendEvent: [NSEvent keyEventWithType: NSKeyDown
+                                      location: NSZeroPoint
+                                 modifierFlags: NSAlternateKeyMask
+                                     timestamp: NSTimeIntervalSince1970
+                                  windowNumber: [self windowNumber]
+                                       context: [self graphicsContext]
+                                    characters: @""
+                   charactersIgnoringModifiers: @""
+                                     isARepeat: NO
+                                       keyCode: NSKEY_RIGHT]];
 }
 
 - (IBAction) performZoom: (id) sender
@@ -366,20 +498,16 @@ char *winfilters[] =
     if (filter != FILTER_ALL)
     {
         NSArray * filterTypes = [NSArray arrayWithObject: [NSString stringWithCString: winfilters[filter]
-                                                                             encoding: NSASCIIStringEncoding]];        
+                                                                             encoding: NSUTF8StringEncoding]];        
         [openDlg setAllowedFileTypes: filterTypes];
         [openDlg setAllowsOtherFileTypes: NO];
-        result = [openDlg runModalForDirectory: NULL file: NULL types: filterTypes];
-    }
-    else
-    {
-        result = [openDlg runModal];
     }
+    result = [openDlg runModal];
 
     if (result == NSFileHandlingPanelOKButton)
-        return [openDlg filename];
+        return [[openDlg URL] path];
 
-    return NULL;
+    return nil;
 }
 
 - (NSString *) saveFileDialog: (NSString *) prompt
@@ -395,7 +523,7 @@ char *winfilters[] =
     if (filter != FILTER_ALL)
     {
         NSArray * filterTypes = [NSArray arrayWithObject: [NSString stringWithCString: winfilters[filter]
-                                                                             encoding: NSASCIIStringEncoding]];        
+                                                                             encoding: NSUTF8StringEncoding]];        
         [saveDlg setAllowedFileTypes: filterTypes];
         [saveDlg setAllowsOtherFileTypes: NO];
     }
@@ -403,9 +531,9 @@ char *winfilters[] =
     result = [saveDlg runModal];
 
     if (result == NSFileHandlingPanelOKButton)
-        return [saveDlg filename];
+        return [[saveDlg URL] path];
 
-    return NULL;
+    return nil;
 }
 
 - (pid_t) retrieveID
@@ -424,7 +552,8 @@ char *winfilters[] =
 
 @end
 
-@interface GargoyleApp : NSWindowController //<NSWindowDelegate>
+@interface GargoyleApp : NSWindowController
+                         <GargoyleApp, NSApplicationDelegate, NSWindowDelegate>
 {
     BOOL openedFirstGame;
     NSMutableDictionary * windows;
@@ -455,12 +584,7 @@ char *winfilters[] =
     /* set environment variable */
     NSString * nsResources = [[NSBundle mainBundle] resourcePath];
 
-    int size = [nsResources length];
-    CFStringGetBytes((CFStringRef) nsResources, CFRangeMake(0, size),
-                     kCFStringEncodingASCII, 0, FALSE,
-                     etc, MaxBuffer, NULL);
-    int bounds = size < MaxBuffer ? size : MaxBuffer;
-    etc[bounds] = '\0';
+    [nsResources getCString: etc maxLength: sizeof etc encoding: NSUTF8StringEncoding];
 
     setenv("GARGLK_INI", etc, TRUE);
 
@@ -498,31 +622,30 @@ char *winfilters[] =
     [window center];
     [window setReleasedWhenClosed: YES];
     [window setDelegate: self];
-    [windows setObject: window forKey: [NSString stringWithFormat: @"%04x", processID]];
+
+    [windows setObject: window forKey: [NSNumber numberWithInt: processID]];
 
     return ([window isVisible]);
 }
 
 - (NSEvent *) getWindowEvent: (pid_t) processID
 {
-    id storedWindow = [windows objectForKey: [NSString stringWithFormat: @"%04x", processID]];
+    GargoyleWindow * window = [windows objectForKey: [NSNumber numberWithInt: processID]];
 
-    if (storedWindow)
+    if (window)
     {
-        GargoyleWindow * window = (GargoyleWindow *) storedWindow;
         return [window retrieveEvent];
     }
 
-    return NULL;
+    return nil;
 }
 
 - (NSRect) getWindowSize: (pid_t) processID
 {
-    id storedWindow = [windows objectForKey: [NSString stringWithFormat: @"%04x", processID]];
+    GargoyleWindow * window = [windows objectForKey: [NSNumber numberWithInt: processID]];
 
-    if (storedWindow)
+    if (window)
     {
-        GargoyleWindow * window = (GargoyleWindow *) storedWindow;
         return [[window contentView] bounds];
     }
 
@@ -531,24 +654,22 @@ char *winfilters[] =
 
 - (NSString *) getWindowCharString: (pid_t) processID
 {
-    id storedWindow = [windows objectForKey: [NSString stringWithFormat: @"%04x", processID]];
+    GargoyleWindow * window = [windows objectForKey: [NSNumber numberWithInt: processID]];
 
-    if (storedWindow)
+    if (window)
     {
-        GargoyleWindow * window = (GargoyleWindow *) storedWindow;
         return [window retrieveChars];
     }
 
-    return NULL;
+    return nil;
 }
 
 - (BOOL) clearWindowCharString: (pid_t) processID
 {
-    id storedWindow = [windows objectForKey: [NSString stringWithFormat: @"%04x", processID]];
+    GargoyleWindow * window = [windows objectForKey: [NSNumber numberWithInt: processID]];
 
-    if (storedWindow)
+    if (window)
     {
-        GargoyleWindow * window = (GargoyleWindow *) storedWindow;
         [window clearChars];
         return YES;
     }
@@ -559,11 +680,10 @@ char *winfilters[] =
 - (BOOL) setWindow: (pid_t) processID
         charString: (NSEvent *) event;
 {
-    id storedWindow = [windows objectForKey: [NSString stringWithFormat: @"%04x", processID]];
+    GargoyleWindow * window = [windows objectForKey: [NSNumber numberWithInt: processID]];
 
-    if (storedWindow)
+    if (window)
     {
-        GargoyleWindow * window = (GargoyleWindow *) storedWindow;
         [window sendChars: event];
         return YES;
     }
@@ -574,11 +694,10 @@ char *winfilters[] =
 - (BOOL) setWindow: (pid_t) processID
              title: (NSString *) title;
 {
-    id storedWindow = [windows objectForKey: [NSString stringWithFormat: @"%04x", processID]];
+    GargoyleWindow * window = [windows objectForKey: [NSNumber numberWithInt: processID]];
 
-    if (storedWindow)
+    if (window)
     {
-        GargoyleWindow * window = (GargoyleWindow *) storedWindow;
         [window setTitle: title];
         return YES;
     }
@@ -591,12 +710,10 @@ char *winfilters[] =
              width: (unsigned int) width
             height: (unsigned int) height;
 {
-    id storedWindow = [windows objectForKey: [NSString stringWithFormat: @"%04x", processID]];
+    GargoyleWindow * window = [windows objectForKey: [NSNumber numberWithInt: processID]];
 
-    if (storedWindow)
+    if (window)
     {
-        GargoyleWindow * window = (GargoyleWindow *) storedWindow;
-
         [[window contentView] addFrame: frame
                                  width: width
                                 height: height];
@@ -610,11 +727,10 @@ char *winfilters[] =
 
 - (void) closeWindow: (pid_t) processID
 {
-    id storedWindow = [windows objectForKey: [NSString stringWithFormat: @"%04x", processID]];
+    GargoyleWindow * window = [windows objectForKey: [NSNumber numberWithInt: processID]];
 
-    if (storedWindow)
+    if (window)
     {
-        GargoyleWindow * window = (GargoyleWindow *) storedWindow;
         [window performClose:self];
     }
 
@@ -624,36 +740,34 @@ char *winfilters[] =
                          prompt: (NSString *) prompt
                          filter: (unsigned int) filter
 {
-    id storedWindow = [windows objectForKey: [NSString stringWithFormat: @"%04x", processID]];
+    GargoyleWindow * window = [windows objectForKey: [NSNumber numberWithInt: processID]];
 
-    if (storedWindow)
+    if (window)
     {
-        GargoyleWindow * window = (GargoyleWindow *) storedWindow;
         return [window openFileDialog: prompt fileFilter: filter];
     }
 
-    return NULL;
+    return nil;
 }
 
 - (NSString *) saveWindowDialog: (pid_t) processID
                          prompt: (NSString *) prompt
                          filter: (unsigned int) filter
 {
-    id storedWindow = [windows objectForKey: [NSString stringWithFormat: @"%04x", processID]];
+    GargoyleWindow * window = [windows objectForKey: [NSNumber numberWithInt: processID]];
 
-    if (storedWindow)
+    if (window)
     {
-        GargoyleWindow * window = (GargoyleWindow *) storedWindow;
         return [window saveFileDialog: prompt fileFilter: filter];
     }
 
-    return NULL;
+    return nil;
 }
 
 - (void) abortWindowDialog: (pid_t) processID
                     prompt: (NSString *) prompt
 {
-    NSRunAlertPanel(@"Fatal error", prompt, NULL, NULL, NULL);
+    NSRunAlertPanel(@"Fatal error", @"%@", nil, nil, nil, prompt);
 }
 
 #define kArrowCursor 1
@@ -690,7 +804,7 @@ char *winfilters[] =
     [openDlg setCanChooseFiles: YES];
     [openDlg setCanChooseDirectories: NO];
     [openDlg setAllowsMultipleSelection: NO];
-    [openDlg setTitle: [NSString stringWithCString: AppName encoding: NSASCIIStringEncoding]];
+    [openDlg setTitle: [NSString stringWithCString: AppName encoding: NSUTF8StringEncoding]];
     
     NSMutableArray *filterTypes = [NSMutableArray arrayWithCapacity:100];
     
@@ -705,10 +819,10 @@ char *winfilters[] =
 
     [openDlg setAllowedFileTypes: filterTypes];
     [openDlg setAllowsOtherFileTypes: NO];
-    result = [openDlg runModalForDirectory: NULL file: NULL types: filterTypes];
+    result = [openDlg runModal];
 
     if (result == NSFileHandlingPanelOKButton)
-        return [self launchFile:[openDlg filename]];
+        return [self launchFile: [[openDlg URL] path]];
 
     return NO;
 }
@@ -722,12 +836,8 @@ char *winfilters[] =
     winpath(dir);
 
     /* get story file */
-    int size = [file length];
-    CFStringGetBytes((CFStringRef) file, CFRangeMake(0, size),
-                     kCFStringEncodingASCII, 0, FALSE,
-                     buf, MaxBuffer, NULL);
-    int bounds = size < MaxBuffer ? size : MaxBuffer;
-    buf[bounds] = '\0';
+    if (![file getCString: buf maxLength: sizeof buf encoding: NSUTF8StringEncoding])
+        return NO;
 
     /* run story file */
     int ran = rungame(dir, buf);
@@ -755,7 +865,7 @@ char *winfilters[] =
     return [self launchFile:file];
 }
 
-- (BOOL) application: (NSApplication *) theApplication openFiles: (NSArray *) files
+- (void) application: (NSApplication *) theApplication openFiles: (NSArray *) files
 {
     openedFirstGame = YES;
 
@@ -766,8 +876,6 @@ char *winfilters[] =
     {
         result = result && [self launchFile:[files objectAtIndex: i]];
     }
-
-    return result;
 }
 
 - (void) openURL: (NSAppleEventDescriptor *) event withReplyEvent: (NSAppleEventDescriptor *) reply
@@ -777,12 +885,12 @@ char *winfilters[] =
     if (urlParts && [urlParts count] == 2)
     {
         openedFirstGame = YES;
-        NSString * game = [[urlParts objectAtIndex: 1] stringByReplacingPercentEscapesUsingEncoding: NSASCIIStringEncoding];
+        NSString * game = [[urlParts objectAtIndex: 1] stringByReplacingPercentEscapesUsingEncoding:NSUTF8StringEncoding];
 
         if ([[NSFileManager defaultManager] fileExistsAtPath: game] == YES)
             [self launchFile: game];
         else
-            NSRunAlertPanel(@"Could not open URL path:", game, NULL, NULL, NULL);
+            NSRunAlertPanel(@"Could not open URL path:", @"%@", nil, nil, nil, game);
     }
 
 }
@@ -790,20 +898,15 @@ char *winfilters[] =
 - (BOOL) windowShouldClose: (id) sender
 {
     GargoyleWindow * window = sender;
-    [windows removeObjectForKey: [NSString stringWithFormat: @"%04x", [window retrieveID]]];
+    [windows removeObjectForKey: [NSNumber numberWithInt: [window retrieveID]]];
     [window quit];
     return YES;
 }
 
 - (NSApplicationTerminateReply) applicationShouldTerminate: (NSApplication *) sender
 {
-    NSEnumerator * windowList = [windows objectEnumerator];
-
-    id storedWindow;
-
-    while (storedWindow = [windowList nextObject])
+    for (GargoyleWindow * window in windows)
     {
-        GargoyleWindow * window = (GargoyleWindow *) storedWindow;
         [window quit];
     }
 
@@ -837,8 +940,8 @@ char *winfilters[] =
 
 void winmsg(const char *msg)
 {
-    NSString * nsMsg = [NSString stringWithCString: msg encoding: NSASCIIStringEncoding];
-    NSRunAlertPanel(@"Fatal error", nsMsg, NULL, NULL, NULL);
+    NSString * nsMsg = [NSString stringWithCString: msg encoding: NSUTF8StringEncoding];
+    NSRunAlertPanel(@"Fatal error", @"%@", nil, nil, nil, nsMsg);
 }
 
 void winpath(char *buffer)
@@ -867,7 +970,7 @@ int winexec(const char *cmd, char **args)
     NSTask * proc = [[NSTask alloc] init];
 
     /* prepare interpreter path */
-    NSArray * nsArray = [[NSString stringWithCString: cmd encoding: NSASCIIStringEncoding] componentsSeparatedByString: @"/"];
+    NSArray * nsArray = [[NSString stringWithCString: cmd encoding: NSUTF8StringEncoding] componentsSeparatedByString: @"/"];
     NSString * nsTerp = [nsArray objectAtIndex: [nsArray count] - 1];
     NSString * nsCmd = [NSString stringWithFormat: @"%@/%@", [[NSBundle mainBundle] builtInPlugInsPath], nsTerp];
 
@@ -879,10 +982,10 @@ int winexec(const char *cmd, char **args)
     [nsEnv setObject: [NSString stringWithFormat: @"com.googlecode.garglk-%04x", getpid()] forKey: @"GargoyleApp"];
 
     if (args[1])
-        [nsArgs addObject: [[NSString alloc] initWithCString: args[1] encoding: NSASCIIStringEncoding]];
+        [nsArgs addObject: [[NSString alloc] initWithCString: args[1] encoding: NSUTF8StringEncoding]];
 
     if (args[2])
-        [nsArgs addObject: [[NSString alloc] initWithCString: args[2] encoding: NSASCIIStringEncoding]];
+        [nsArgs addObject: [[NSString alloc] initWithCString: args[2] encoding: NSUTF8StringEncoding]];
 
     if ([nsCmd length] && [nsArgs count])
     {
