@@ -105,12 +105,13 @@ void winexit(void)
     exit(0);
 }
 
-void winchoosefile(char *prompt, char *buf, int len, int filter, GtkFileChooserAction action, const char *button)
+static void winchoosefile(char *prompt, char *buf, int len, int filter, GtkFileChooserAction action, const char *button)
 {
     GtkWidget *filedlog = gtk_file_chooser_dialog_new(prompt, NULL, action,
                                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                                       button, GTK_RESPONSE_ACCEPT,
                                                       NULL);
+    char *curdir;
 
     if (filter != FILTER_ALL)
     {
@@ -152,8 +153,12 @@ void winchoosefile(char *prompt, char *buf, int len, int filter, GtkFileChooserA
     else
         strcpy(buf, "");
 
-    strcpy(filepath, gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(filedlog)));
-    fileselect = TRUE;
+    curdir = gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(filedlog));
+    if (curdir != NULL && strlen(curdir) < sizeof(filepath))
+    {
+        strcpy(filepath, curdir);
+        fileselect = TRUE;
+    }
 
     gtk_widget_destroy(filedlog);
     filedlog = NULL;
@@ -224,10 +229,10 @@ void winclipstore(glui32 *text, int len)
 
     /* null-terminated string */
     cliptext[k] = '\0';
-    cliplen = k + 1;
+    cliplen = k;
 }
 
-void winclipsend(int source)
+static void winclipsend(int source)
 {
     if (!cliplen)
         return;
@@ -249,7 +254,7 @@ void winclipsend(int source)
     }
 }
 
-void winclipreceive(int source)
+static void winclipreceive(int source)
 {
     gchar *gptr;
     int glen, i;
@@ -423,11 +428,18 @@ static void onkeydown(GtkWidget *widget, GdkEventKey *event, void *data)
         switch(key)
         {
             case GDK_a: case GDK_A: gli_input_handle_key(keycode_Home); break;
+            case GDK_b: case GDK_B: gli_input_handle_key(keycode_Left); break;
             case GDK_c: case GDK_C: winclipsend(CLIPBOARD); break;
+            case GDK_d: case GDK_D: gli_input_handle_key(keycode_Erase); break;
             case GDK_e: case GDK_E: gli_input_handle_key(keycode_End); break;
+            case GDK_f: case GDK_F: gli_input_handle_key(keycode_Right); break;
+            case GDK_n: case GDK_N: gli_input_handle_key(keycode_Down); break;
+            case GDK_p: case GDK_P: gli_input_handle_key(keycode_Up); break;
             case GDK_u: case GDK_U: gli_input_handle_key(keycode_Escape); break;
             case GDK_v: case GDK_V: winclipreceive(CLIPBOARD); break;
             case GDK_x: case GDK_X: winclipsend(CLIPBOARD); break;
+            case GDK_Left: gli_input_handle_key(keycode_SkipWordLeft); break;
+            case GDK_Right: gli_input_handle_key(keycode_SkipWordRight); break;
         }
 
         return;
@@ -554,13 +566,26 @@ void winopen(void)
     g_signal_connect(imcontext, "commit",
         G_CALLBACK(oninput), NULL);
 
+    if (gli_conf_fullscreen)
+    {
+        GdkScreen *screen = gdk_screen_get_default();
+        GdkRectangle monitor;
+        int root_monitor = 0;
+
+        gdk_screen_get_monitor_geometry(screen, root_monitor, &monitor);
+
+        defw = monitor.width;
+        defh = monitor.height;
+        gtk_window_set_decorated(GTK_WINDOW(frame), FALSE);
+        gtk_window_set_position(GTK_WINDOW(frame), GTK_WIN_POS_CENTER);
+        gtk_window_fullscreen(GTK_WINDOW(frame));       
+    }
+
     wintitle();
 
     gtk_window_set_geometry_hints(GTK_WINDOW(frame),
         GTK_WIDGET(frame), &geom,
-        GDK_HINT_MIN_SIZE
-        | GDK_HINT_MAX_SIZE
-        | GDK_HINT_RESIZE_INC
+        GDK_HINT_MIN_SIZE | GDK_HINT_MAX_SIZE
         );
     gtk_window_set_default_size(GTK_WINDOW(frame), defw, defh);
 
