@@ -30,12 +30,7 @@
 #include "garglk.h"
 
 #import "Cocoa/Cocoa.h"
-
-#ifdef __ppc__
-#define ByteOrderUCS4 kCFStringEncodingUTF32
-#else
-#define ByteOrderUCS4 kCFStringEncodingUTF32LE
-#endif
+#import "sysmac.h"
 
 static volatile int gli_event_waiting = FALSE;
 static volatile int gli_mach_allowed = FALSE;
@@ -47,47 +42,6 @@ static volatile int gli_window_alive = TRUE;
 
 void wintick(CFRunLoopTimerRef timer, void *info);
 
-@protocol GargoyleApp
-- (BOOL) initWindow: (pid_t) processID
-              width: (unsigned int) width
-             height: (unsigned int) height;
-
-- (NSEvent *) getWindowEvent: (pid_t) processID;
-
-- (NSRect) getWindowSize: (pid_t) processID;
-
-- (NSString *) getWindowCharString: (pid_t) processID;
-
-- (BOOL) clearWindowCharString: (pid_t) processID;
-
-- (BOOL) setWindow: (pid_t) processID
-        charString: (NSEvent *) event;
-
-- (BOOL) setWindow: (pid_t) processID
-             title: (NSString *) title;
-
-- (BOOL) setWindow: (pid_t) processID
-          contents: (NSData *) frame
-             width: (unsigned int) width
-            height: (unsigned int) height;
-
-- (void) closeWindow: (pid_t) processID;
-
-- (NSString *) openWindowDialog: (pid_t) processID
-                         prompt: (NSString *) prompt
-                         filter: (unsigned int) filter;
-
-- (NSString *) saveWindowDialog: (pid_t) processID
-                         prompt: (NSString *) prompt
-                         filter: (unsigned int) filter;
-
-- (void) abortWindowDialog: (pid_t) processID
-                    prompt: (NSString *) prompt;
-
-- (void) setCursor: (unsigned int) cursor;
-
-@end
-
 @interface GargoyleMonitor : NSObject
 {
     NSRect size;
@@ -200,7 +154,7 @@ void winabort(const char *fmt, ...)
 
     NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
     [gargoyle abortWindowDialog: processID
-                         prompt: [NSString stringWithCString: buf encoding: NSASCIIStringEncoding]];
+                         prompt: [NSString stringWithCString: buf encoding: NSUTF8StringEncoding]];
     [pool drain];
 
     exit(1);
@@ -217,21 +171,14 @@ void winopenfile(char *prompt, char *buf, int len, int filter)
     NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];    
 
     NSString * fileref = [gargoyle openWindowDialog: processID
-                                             prompt: [NSString stringWithCString: prompt encoding: NSASCIIStringEncoding]
+                                             prompt: [NSString stringWithCString: prompt encoding: NSUTF8StringEncoding]
                                              filter: filter];
 
     strcpy(buf, "");
 
     if (fileref)
     {
-        int size = [fileref length];
-
-        CFStringGetBytes((CFStringRef) fileref, CFRangeMake(0, size),
-                         kCFStringEncodingASCII, 0, FALSE,
-                         buf, len, NULL);
-
-        int bounds = size < len ? size : len;
-        buf[bounds] = '\0';
+        [fileref getCString: buf maxLength: len encoding: NSUTF8StringEncoding];
     }
 
     [pool drain];
@@ -242,21 +189,14 @@ void winsavefile(char *prompt, char *buf, int len, int filter)
     NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
 
     NSString * fileref = [gargoyle saveWindowDialog: processID
-                                             prompt: [NSString stringWithCString: prompt encoding: NSASCIIStringEncoding]
+                                             prompt: [NSString stringWithCString: prompt encoding: NSUTF8StringEncoding]
                                              filter: filter];
 
     strcpy(buf, "");
 
     if (fileref)
     {
-        int size = [fileref length];
-
-        CFStringGetBytes((CFStringRef) fileref, CFRangeMake(0, size),
-                         kCFStringEncodingASCII, 0, FALSE,
-                         buf, len, NULL);
-
-        int bounds = size < len ? size : len;
-        buf[bounds] = '\0';
+        [fileref getCString: buf maxLength: len encoding: NSUTF8StringEncoding];
     }
 
     [pool drain];
@@ -269,12 +209,12 @@ void winclipstore(glui32 *text, int len)
 
     if (cliptext) {
         [cliptext release];
-        cliptext = NULL;
+        cliptext = nil;
     }
 
-    cliptext = (NSString *) CFStringCreateWithBytes(kCFAllocatorDefault,
-                                                    (char *) text, (len * 4),
-                                                    ByteOrderUCS4, FALSE);
+    cliptext = [[NSString alloc] initWithBytes: text
+                                        length: (len * sizeof(glui32))
+                                      encoding: UTF32StringEncoding];
 }
 
 void winclipsend(void)
@@ -302,9 +242,11 @@ void winclipreceive(void)
             len = [input length];
             for (i=0; i < len; i++)
             {
-                if (CFStringGetBytes((CFStringRef) input, CFRangeMake(i, 1),
-                                     kCFStringEncodingUTF32, 0, FALSE,
-                                     (char *) &ch, 4, NULL))
+                if ([input getBytes: &ch maxLength: sizeof ch usedLength: NULL
+                           encoding: UTF32StringEncoding
+                            options: 0
+                              range: NSMakeRange(i, 1)
+                     remainingRange: NULL])
                 {
                     switch (ch)
                     {
@@ -330,16 +272,17 @@ void wintitle(void)
 {
     NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
 
-    char buf[256];
+    NSString * story_title = [NSString stringWithCString: gli_story_title encoding: NSUTF8StringEncoding];
+    NSString * story_name = [NSString stringWithCString: gli_story_name encoding: NSUTF8StringEncoding];
+    NSString * program_name = [NSString stringWithCString: gli_program_name encoding: NSUTF8StringEncoding];
 
-    if (strlen(gli_story_title))
-        sprintf(buf, "%s", gli_story_title);
-    else if (strlen(gli_story_name))
-        sprintf(buf, "%s - %s", gli_story_name, gli_program_name);
+    NSString * title = nil;
+    if ([story_title length])
+        title = story_title;
+    else if ([story_name length])
+        title = [NSString stringWithFormat: @"%@ - %@", story_name, program_name];
     else
-        sprintf(buf, "%s", gli_program_name);
-
-    NSString * title = [NSString stringWithCString: buf encoding: NSASCIIStringEncoding];
+        title = program_name;
 
     [gargoyle setWindow: processID
                   title: title];
@@ -479,36 +422,6 @@ void winrefresh(void)
     gli_refresh_needed = !refreshed;
 }
 
-#define NSKEY_LEFT      0x7b
-#define NSKEY_RIGHT     0x7c
-#define NSKEY_DOWN      0x7d
-#define NSKEY_UP        0x7e
-
-#define NSKEY_X         0x07
-#define NSKEY_C         0x08
-#define NSKEY_V         0x09
-
-#define NSKEY_PGUP      0x74
-#define NSKEY_PGDN      0x79
-#define NSKEY_HOME      0x73
-#define NSKEY_END       0x77
-#define NSKEY_DEL       0x75
-#define NSKEY_BACK      0x33
-#define NSKEY_ESC       0x35
-
-#define NSKEY_F1        0x7a
-#define NSKEY_F2        0x78
-#define NSKEY_F3        0x63
-#define NSKEY_F4        0x76
-#define NSKEY_F5        0x60
-#define NSKEY_F6        0x61
-#define NSKEY_F7        0x62
-#define NSKEY_F8        0x64
-#define NSKEY_F9        0x65
-#define NSKEY_F10       0x6d
-#define NSKEY_F11       0x67
-#define NSKEY_F12       0x6f
-
 void winkey(NSEvent *evt)
 {
     /* queue a screen refresh */
@@ -517,10 +430,21 @@ void winkey(NSEvent *evt)
     /* check for arrow keys */
     if ([evt modifierFlags] & NSFunctionKeyMask)
     {
-        /* modified keys for scrolling */
-        if ([evt modifierFlags] & NSCommandKeyMask
-            || [evt modifierFlags] & NSAlternateKeyMask
-            || [evt modifierFlags] & NSControlKeyMask)
+        /* alt/option modified key */
+        if ([evt modifierFlags] & NSAlternateKeyMask)
+        {
+            switch ([evt keyCode])
+            {
+                case NSKEY_LEFT  : gli_input_handle_key(keycode_SkipWordLeft);  return;
+                case NSKEY_RIGHT : gli_input_handle_key(keycode_SkipWordRight); return;
+                case NSKEY_DOWN  : gli_input_handle_key(keycode_PageDown);      return;
+                case NSKEY_UP    : gli_input_handle_key(keycode_PageUp);        return;
+                default: break;
+            }
+        }
+
+        /* command modified key */
+        if ([evt modifierFlags] & NSCommandKeyMask)
         {
             switch ([evt keyCode])
             {
@@ -532,17 +456,14 @@ void winkey(NSEvent *evt)
             }
         }
 
-        /* unmodified keys for line editing */
-        else
+        /* unmodified key for line editing */
+        switch ([evt keyCode])
         {
-            switch ([evt keyCode])
-            {
-                case NSKEY_LEFT  : gli_input_handle_key(keycode_Left);  return;
-                case NSKEY_RIGHT : gli_input_handle_key(keycode_Right); return;
-                case NSKEY_DOWN  : gli_input_handle_key(keycode_Down);  return;
-                case NSKEY_UP    : gli_input_handle_key(keycode_Up);    return;
-                default: break;
-            }
+            case NSKEY_LEFT  : gli_input_handle_key(keycode_Left);  return;
+            case NSKEY_RIGHT : gli_input_handle_key(keycode_Right); return;
+            case NSKEY_DOWN  : gli_input_handle_key(keycode_Down);  return;
+            case NSKEY_UP    : gli_input_handle_key(keycode_Up);    return;
+            default: break;
         }
     }
 
@@ -601,10 +522,12 @@ void winkey(NSEvent *evt)
 
     /* convert character to UTF-32 value */
     glui32 ch;
-    if (CFStringGetBytes((CFStringRef) evt_char,
-                         CFRangeMake(0, [evt_char length]),
-                         kCFStringEncodingUTF32, 0, FALSE,
-                         (char *)&ch, 4, NULL)) {
+    if ([evt_char getBytes: &ch maxLength: sizeof ch usedLength: NULL
+                  encoding: UTF32StringEncoding
+                   options: 0
+                     range: NSMakeRange(0, [evt_char length])
+            remainingRange: NULL])
+    {
         switch (ch)
         {
             case '\n': gli_input_handle_key(keycode_Return); break;
