@@ -24,6 +24,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <stdbool.h>
 #include "glk.h"
 #include "garglk.h"
 
@@ -85,6 +86,7 @@ window_textbuffer_t *win_textbuffer_create(window_t *win)
     dwin->height = -1;
 
     dwin->inbuf = NULL;
+    dwin->inunicode = FALSE;
     dwin->line_terminators = NULL;
     dwin->echo_line_input = TRUE;
 
@@ -124,10 +126,15 @@ window_textbuffer_t *win_textbuffer_create(window_t *win)
 
 void win_textbuffer_destroy(window_textbuffer_t *dwin)
 {
+    int i;
+
     if (dwin->inbuf)
     {
         if (gli_unregister_arr)
-            (*gli_unregister_arr)(dwin->inbuf, dwin->inmax, "&+#!Cn", dwin->inarrayrock);
+        {
+            char *typedesc = (dwin->inunicode ? "&+#!Iu" : "&+#!Cn");
+            (*gli_unregister_arr)(dwin->inbuf, dwin->inmax, typedesc, dwin->inarrayrock);
+        }
         dwin->inbuf = NULL;
     }
 
@@ -139,6 +146,11 @@ void win_textbuffer_destroy(window_textbuffer_t *dwin)
     if (dwin->line_terminators)
         free(dwin->line_terminators);
 
+    for (i = 0; i < dwin->scrollback; i++) {
+        gli_picture_decrement(dwin->lines[i].lpic);
+        gli_picture_decrement(dwin->lines[i].rpic);
+    }
+
     free(dwin->lines);
     free(dwin);
 }
@@ -161,12 +173,20 @@ static void reflow(window_t *win)
     attr_t *attrbuf = malloc(sizeof(attr_t) * SCROLLBACK * TBLINELEN);
     glui32 *charbuf = malloc(sizeof(glui32) * SCROLLBACK * TBLINELEN);
     int *alignbuf = malloc(sizeof(int) * SCROLLBACK);
-    picture_t **pictbuf = malloc(sizeof(size_t) * SCROLLBACK);
+    picture_t **pictbuf = malloc(sizeof(picture_t *) * SCROLLBACK);
     glui32 *hyperbuf = malloc(sizeof(glui32) * SCROLLBACK);
     int *offsetbuf = malloc(sizeof(int) * SCROLLBACK);
 
     if (!attrbuf || !charbuf || !alignbuf || !pictbuf || !hyperbuf || !offsetbuf)
+    {
+        free(attrbuf);
+        free(charbuf);
+        free(alignbuf);
+        free(pictbuf);
+        free(hyperbuf);
+        free(offsetbuf);
         return;
+    }
 
     /* copy text to temp buffers */
 
@@ -187,6 +207,7 @@ static void reflow(window_t *win)
             offsetbuf[x] = p;
             alignbuf[x] = imagealign_MarginLeft;
             pictbuf[x] = dwin->lines[k].lpic;
+            gli_picture_increment(pictbuf[x]);
             hyperbuf[x] = dwin->lines[k].lhyper;
             x++;
         }
@@ -196,6 +217,7 @@ static void reflow(window_t *win)
             offsetbuf[x] = p;
             alignbuf[x] = imagealign_MarginRight;
             pictbuf[x] = dwin->lines[k].rpic;
+            gli_picture_increment(pictbuf[x]);
             hyperbuf[x] = dwin->lines[k].rhyper;
             x++;
         }
@@ -522,9 +544,12 @@ void win_textbuffer_redraw(window_t *win)
                     dwin->copypos++;
                 }
             }
-            /* add newline to copy buffer */
-            dwin->copybuf[dwin->copypos] = '\n';
-            dwin->copypos++;
+            /* add newline if we reach the end of the line */
+            if (ln->len == 0 || ln->len == (rsc+1))
+            {
+                dwin->copybuf[dwin->copypos] = '\n';
+                dwin->copypos++;
+            }
         }
 
         /* clear any stored hyperlink coordinates */
@@ -939,6 +964,29 @@ static void put_text_uni(window_textbuffer_t *dwin, glui32 *buf, int len, int po
     touch(dwin, 0);
 }
 
+/* Return true if a following quotation mark should be an opening mark,
+ * false if it should be a closing mark. Opening quotation marks will
+ * appear following an open parenthesis, open square bracket, or
+ * whitespace.
+ */
+static bool leftquote(uint32_t c)
+{
+    switch(c)
+    {
+        case '(': case '[':
+
+        /* The following are Unicode characters in the "Separator, Space" category. */
+        case 0x0020: case 0x00a0: case 0x1680: case 0x2000:
+        case 0x2001: case 0x2002: case 0x2003: case 0x2004:
+        case 0x2005: case 0x2006: case 0x2007: case 0x2008:
+        case 0x2009: case 0x200a: case 0x202f: case 0x205f:
+        case 0x3000:
+            return true;
+        default:
+            return false;
+    }
+}
+
 void win_textbuffer_putchar_uni(window_t *win, glui32 ch)
 {
     window_textbuffer_t *dwin = win->data;
@@ -951,9 +999,7 @@ void win_textbuffer_putchar_uni(window_t *win, glui32 ch)
     int linelen;
     unsigned char *color;
 
-#ifdef USETTS
-    { char b[1]; b[0] = ch; gli_speak_tts(b, 1, 0); }
-#endif
+    gli_tts_speak(&ch, 1);
 
     pw = (win->bbox.x1 - win->bbox.x0 - gli_tmarginx * 2 - gli_scroll_width) * GLI_SUBPIX;
     pw = pw - 2 * SLOP - dwin->radjw - dwin->ladjw;
@@ -972,11 +1018,10 @@ void win_textbuffer_putchar_uni(window_t *win, glui32 ch)
 
     if (gli_conf_quotes)
     {
-#define LEFTQUOTE(c)	((c) == ' ' || (c) == '(' || (c) == '[')
         /* fails for 'tis a wonderful day in the '80s */
         if (gli_conf_quotes > 1 && ch == '\'')
         {
-            if (dwin->numchars == 0 || LEFTQUOTE(dwin->chars[dwin->numchars-1]))
+            if (dwin->numchars == 0 || leftquote(dwin->chars[dwin->numchars-1]))
                 ch = UNI_LSQUO;
         }
 
@@ -988,15 +1033,14 @@ void win_textbuffer_putchar_uni(window_t *win, glui32 ch)
 
         if (ch == '"')
         {
-            if (dwin->numchars == 0 || LEFTQUOTE(dwin->chars[dwin->numchars-1]))
+            if (dwin->numchars == 0 || leftquote(dwin->chars[dwin->numchars-1]))
                 ch = UNI_LDQUO;
             else
                 ch = UNI_RDQUO;
         }
-#undef LEFTQUOTE
     }
 
-    if (gli_conf_quotes && win->attr.style != style_Preformatted)
+    if (gli_conf_dashes && win->attr.style != style_Preformatted)
     {
         if (ch == '-')
         {
@@ -1004,7 +1048,10 @@ void win_textbuffer_putchar_uni(window_t *win, glui32 ch)
             if (dwin->dashed == 2)
             {
                 dwin->numchars--;
-                ch = UNI_NDASH;
+                if (gli_conf_dashes == 2)
+                    ch = UNI_NDASH;
+                else
+                    ch = UNI_MDASH;
             }
             if (dwin->dashed == 3)
             {
@@ -1126,8 +1173,12 @@ void win_textbuffer_clear(window_t *win)
     for (i = 0; i < dwin->scrollback; i++)
     {
         dwin->lines[i].len = 0;
+
+        gli_picture_decrement(dwin->lines[i].lpic);
         dwin->lines[i].lpic = 0;
+        gli_picture_decrement(dwin->lines[i].rpic);
         dwin->lines[i].rpic = 0;
+
         dwin->lines[i].lhyper = 0;
         dwin->lines[i].rhyper = 0;
         dwin->lines[i].lm = 0;
@@ -1158,9 +1209,7 @@ void win_textbuffer_init_line(window_t *win, char *buf, int maxlen, int initlen)
     window_textbuffer_t *dwin = win->data;
     int pw;
 
-#ifdef USETTS
-    gli_speak_tts("\n", 1, 0);
-#endif
+    gli_tts_flush();
 
     /* because '>' prompt is ugly without extra space */
     if (dwin->numchars && dwin->chars[dwin->numchars-1] == '>')
@@ -1177,6 +1226,7 @@ void win_textbuffer_init_line(window_t *win, char *buf, int maxlen, int initlen)
     //dwin->lastseen = 0;
 
     dwin->inbuf = buf;
+    dwin->inunicode = FALSE;
     dwin->inmax = maxlen;
     dwin->infence = dwin->numchars;
     dwin->incurs = dwin->numchars;
@@ -1205,7 +1255,7 @@ void win_textbuffer_init_line(window_t *win, char *buf, int maxlen, int initlen)
     }
 
     if (gli_register_arr)
-        dwin->inarrayrock = (*gli_register_arr)(buf, maxlen, "&+#!Cn");
+        dwin->inarrayrock = (*gli_register_arr)(dwin->inbuf, maxlen, "&+#!Cn");
 }
 
 void win_textbuffer_init_line_uni(window_t *win, glui32 *buf, int maxlen, int initlen)
@@ -1213,9 +1263,7 @@ void win_textbuffer_init_line_uni(window_t *win, glui32 *buf, int maxlen, int in
     window_textbuffer_t *dwin = win->data;
     int pw;
 
-#ifdef USETTS
-    gli_speak_tts("\n", 1, 0);
-#endif
+    gli_tts_flush();
 
     /* because '>' prompt is ugly without extra space */
     if (dwin->numchars && dwin->chars[dwin->numchars-1] == '>')
@@ -1232,6 +1280,7 @@ void win_textbuffer_init_line_uni(window_t *win, glui32 *buf, int maxlen, int in
     //dwin->lastseen = 0;
 
     dwin->inbuf = buf;
+    dwin->inunicode = TRUE;
     dwin->inmax = maxlen;
     dwin->infence = dwin->numchars;
     dwin->incurs = dwin->numchars;
@@ -1260,7 +1309,7 @@ void win_textbuffer_init_line_uni(window_t *win, glui32 *buf, int maxlen, int in
     }
 
     if (gli_register_arr)
-        dwin->inarrayrock = (*gli_register_arr)(buf, maxlen, "&+#!Iu");
+        dwin->inarrayrock = (*gli_register_arr)(dwin->inbuf, maxlen, "&+#!Iu");
 }
 
 /* Abort line input, storing whatever's been typed so far. */
@@ -1271,8 +1320,7 @@ void win_textbuffer_cancel_line(window_t *win, event_t *ev)
     int ix;
     int len;
     void *inbuf;
-    int inmax;
-    int unicode = win->line_request_uni;
+    int inmax, inunicode;
 
     if (!dwin->inbuf)
         return;
@@ -1280,15 +1328,16 @@ void win_textbuffer_cancel_line(window_t *win, event_t *ev)
     inbuf = dwin->inbuf;
     inmax = dwin->inmax;
     inarrayrock = dwin->inarrayrock;
+    inunicode = dwin->inunicode;
 
     len = dwin->numchars - dwin->infence;
-    if (win->echostr) 
+    if (win->echostr)
         gli_stream_echo_line_uni(win->echostr, dwin->chars + dwin->infence, len);
 
     if (len > inmax)
         len = inmax;
 
-    if (!unicode)
+    if (!inunicode)
     {
         for (ix=0; ix<len; ix++)
         {
@@ -1332,16 +1381,20 @@ void win_textbuffer_cancel_line(window_t *win, event_t *ev)
     }
 
     if (gli_unregister_arr)
-        (*gli_unregister_arr)(inbuf, inmax, unicode ? "&+#!Iu" : "&+#!Cn", inarrayrock);
+    {
+        char *typedesc = (inunicode ? "&+#!Iu" : "&+#!Cn");
+        (*gli_unregister_arr)(inbuf, inmax, typedesc, inarrayrock);
+    }
 }
 
 /* Keybinding functions. */
 
 /* Any key, when text buffer is scrolled. */
-void gcmd_accept_scroll(window_t *win, glui32 arg)
+int gcmd_accept_scroll(window_t *win, glui32 arg)
 {
     window_textbuffer_t *dwin = win->data;
     int pageht = dwin->height - 2;        /* 1 for prompt, 1 for overlap */
+    int startpos = dwin->scrollpos;
 
     switch (arg)
     {
@@ -1360,9 +1413,11 @@ void gcmd_accept_scroll(window_t *win, glui32 arg)
             break;
         case keycode_MouseWheelUp:
             dwin->scrollpos += 3;
+            startpos = TRUE;
             break;
         case keycode_MouseWheelDown:
             dwin->scrollpos -= 3;
+            startpos = TRUE;
             break;
         case ' ':
         case keycode_PageDown:
@@ -1379,6 +1434,8 @@ void gcmd_accept_scroll(window_t *win, glui32 arg)
     if (dwin->scrollpos < 0)
         dwin->scrollpos = 0;
     touchscroll(dwin);
+
+    return (startpos || dwin->scrollpos);
 }
 
 /* Any key, during character input. Ends character input. */
@@ -1404,18 +1461,13 @@ void gcmd_buffer_accept_readchar(window_t *win, glui32 arg)
             key = keycode_Delete;
             break;
         case keycode_MouseWheelUp:
-            key = keycode_PageUp;
-            break;
         case keycode_MouseWheelDown:
-            key = keycode_PageDown;
-            break;
+            return;
         default:
             key = arg;
     }
 
-#ifdef USETTS
-    gli_speak_tts("", 0, 1);
-#endif
+    gli_tts_purge();
 
     if (key > 0xff && key < (0xffffffff - keycode_MAXVAL + 1))
     {
@@ -1435,10 +1487,9 @@ static void acceptline(window_t *win, glui32 keycode)
     int len, olen;
     void *inbuf;
     glui32 *s, *o;
-    int inmax;
+    int inmax, inunicode;
     gidispatch_rock_t inarrayrock;
     window_textbuffer_t *dwin = win->data;
-    int unicode = win->line_request_uni;
 
     if (!dwin->inbuf)
         return;
@@ -1446,57 +1497,51 @@ static void acceptline(window_t *win, glui32 keycode)
     inbuf = dwin->inbuf;
     inmax = dwin->inmax;
     inarrayrock = dwin->inarrayrock;
+    inunicode = dwin->inunicode;
 
     len = dwin->numchars - dwin->infence;
-    if (win->echostr) 
+    if (win->echostr)
         gli_stream_echo_line_uni(win->echostr, dwin->chars+dwin->infence, len);
 
-#ifdef USETTS
-    gli_speak_tts(dwin->chars+dwin->infence, len, 1);
-#endif
+    gli_tts_purge();
+    if (gli_conf_speak_input)
+    {
+        gli_tts_speak(dwin->chars + dwin->infence, len);
+        gli_tts_speak((glui32[]){'\n'}, 1);
+    }
 
-    /* Store in history. */
+    /*
+     * Store in history.
+     * The history is a ring buffer, with historypresent being the index of the most recent
+     * element and historyfirst the index of the oldest element.
+     * A history entry should not repeat the string from the entry before it.
+     */
     if (len)
     {
-        s = malloc((len + 1) * 4);
-        memcpy(s, dwin->chars + dwin->infence, len * 4);
+        s = malloc((len + 1) * sizeof(glui32));
+        memcpy(s, dwin->chars + dwin->infence, len * sizeof(glui32));
         s[len] = 0;
 
-        if (dwin->history[dwin->historypresent])
-        {
-            free(dwin->history[dwin->historypresent]);
-            dwin->history[dwin->historypresent] = NULL;
-        }
-
-        if (dwin->history[dwin->historypresent] != dwin->history[dwin->historyfirst] )
-            o = dwin->history[dwin->historypresent - 1];
-        else
-            o = NULL;
+        free(dwin->history[dwin->historypresent]);
+        dwin->history[dwin->historypresent] = NULL;
 
+        o = dwin->history[(dwin->historypresent == 0 ? HISTORYLEN : dwin->historypresent) - 1];
         olen = o ? strlen_uni(o) : 0;
 
         if (len != olen || memcmp(s, o, olen * sizeof(glui32)))
         {
-
             dwin->history[dwin->historypresent] = s;
 
-            dwin->historypresent ++;
-            if (dwin->historypresent >= HISTORYLEN)
-                dwin->historypresent -= HISTORYLEN;
+            dwin->historypresent++;
+            if (dwin->historypresent == HISTORYLEN)
+                dwin->historypresent = 0;
 
             if (dwin->historypresent == dwin->historyfirst)
             {
                 dwin->historyfirst++;
-                if (dwin->historyfirst >= HISTORYLEN)
-                    dwin->historyfirst -= HISTORYLEN;
+                if (dwin->historyfirst == HISTORYLEN)
+                    dwin->historyfirst = 0;
             }
-
-            if (dwin->history[dwin->historypresent])
-            {
-                free(dwin->history[dwin->historypresent]);
-                dwin->history[dwin->historypresent] = NULL;
-            }
-
         }
         else
         {
@@ -1509,7 +1554,7 @@ static void acceptline(window_t *win, glui32 keycode)
     if (len > inmax)
         len = inmax;
 
-    if (!unicode)
+    if (!inunicode)
     {
         for (ix=0; ix<len; ix++)
         {
@@ -1556,7 +1601,10 @@ static void acceptline(window_t *win, glui32 keycode)
     }
 
     if (gli_unregister_arr)
-        (*gli_unregister_arr)(inbuf, inmax, unicode ? "&+#!Iu" : "&+#!Cn", inarrayrock);
+    {
+        char *typedesc = (inunicode ? "&+#!Iu" : "&+#!Cn");
+        (*gli_unregister_arr)(inbuf, inmax, typedesc, inarrayrock);
+    }
 }
 
 /* Any key, during line input. */
@@ -1662,6 +1710,20 @@ void gcmd_buffer_accept_readline(window_t *win, glui32 arg)
             dwin->incurs = dwin->numchars;
             break;
 
+        case keycode_SkipWordLeft:
+            while (dwin->incurs > dwin->infence && dwin->chars[dwin->incurs - 1] == ' ')
+                dwin->incurs--;
+            while (dwin->incurs > dwin->infence && dwin->chars[dwin->incurs - 1] != ' ')
+                dwin->incurs--;
+            break;
+
+        case keycode_SkipWordRight:
+            while (dwin->incurs < dwin->numchars && dwin->chars[dwin->incurs] != ' ')
+                dwin->incurs++;
+            while (dwin->incurs < dwin->numchars && dwin->chars[dwin->incurs] == ' ')
+                dwin->incurs++;
+            break;
+
             /* Delete keys, during line input. */
 
         case keycode_Delete:
@@ -1764,6 +1826,7 @@ glui32 win_textbuffer_draw_picture(window_textbuffer_t *dwin,
 
     hyperlink = dwin->owner->attr.hyper;
 
+    gli_picture_increment(pic);
     error = put_picture(dwin, pic, align, hyperlink);
 
     return error;
