@@ -22,8 +22,11 @@ Modified
 */
 
 #include <string.h>
+#include <stdio.h>
+#include <stdarg.h>
 
 #include <windows.h>
+#include <Wincrypt.h>
 
 #include "std.h"
 #include "os.h"
@@ -44,25 +47,6 @@ static char S_open_file_dir[OSFNMAX];
 extern HINSTANCE oss_G_hinstance;
 
 
-/* ------------------------------------------------------------------------ */
-/*
- *   initialize 
- */
-int os_init(int *argc, char *argv[], const char *prompt,
-            char *buf, int bufsiz)
-{
-    /* nothing to do - just return success */
-    return 0;
-}
-
-/*
- *   uninitialize 
- */
-void os_uninit()
-{
-    /* nothing tod o */
-}
-
 /* ------------------------------------------------------------------------ */
 /*
  *   Set the initial directory for os_askfile dialogs 
@@ -577,6 +561,11 @@ void os_set_save_ext(const char *ext)
     }
 }
 
+const char *os_get_save_ext()
+{
+    return (default_saved_game_ext[0] != '\0' ? default_saved_game_ext : 0);
+}
+
 
 /* ------------------------------------------------------------------------ */
 /*
@@ -1207,3 +1196,4 @@ char *os_strlwr(char *s)
 {
     return strlwr(s);
 }
+
