@@ -46,6 +46,9 @@ Modified
 #include "vmpredef.h"
 #include "vmcset.h"
 #include "vmfilobj.h"
+#include "vmfilnam.h"
+#include "vmnetfil.h"
+#include "vmnet.h"
 
 
 /* ------------------------------------------------------------------------ */
@@ -109,9 +112,14 @@ void CVmBifTIO::say_to_console(VMG_ CVmConsole *console, uint argc)
             /* go display it */
             goto disp_str;
 
+        case VM_LIST:
+            /* convert to a string */
+            str = val.cast_to_string(vmg_ &new_str);
+            goto disp_str;
+
         case VM_INT:
             /* convert it to a string */
-            sprintf(buf + 2, "%ld", val.val.intval);
+            sprintf(buf + 2, "%ld", (long)val.val.intval);
 
             /* set its length */
             len = strlen(buf + 2);
@@ -143,34 +151,32 @@ void CVmBifTIO::say_to_console(VMG_ CVmConsole *console, uint argc)
 
 void CVmBifTIO::logging(VMG_ uint argc)
 {
-    vm_val_t fname_val;
-    int log_type;
+    /* presume success */
+    int ok = TRUE;
     
     /* check arguments */
     check_argc_range(vmg_ argc, 1, 2);
 
-    /* retrieve the filename argument */
-    G_stk->pop(&fname_val);
-
-    /* get the log type argument, if present */
-    log_type = (argc >= 2 ? pop_int_val(vmg0_) : LOG_SCRIPT);
+    /* get the arguments: filename, log type */
+    const vm_val_t *filespec = G_stk->get(0);
+    int log_type = (argc >= 2 ? G_stk->get(1)->num_to_int(vmg0_) : LOG_SCRIPT);
 
     /* 
      *   if they passed us nil, turn off logging; otherwise, start logging
      *   to the filename given by the string 
      */
-    if (fname_val.typ == VM_NIL)
+    if (filespec->typ == VM_NIL)
     {
         /* turn off the appropriate type of logging */
         switch(log_type)
         {
         case LOG_SCRIPT:
-            G_console->close_log_file();
+            G_console->close_log_file(vmg0_);
             break;
 
         case LOG_CMD:
         case LOG_EVENT:
-            G_console->close_command_log();
+            G_console->close_command_log(vmg0_);
             break;
 
         default:
@@ -179,31 +185,33 @@ void CVmBifTIO::logging(VMG_ uint argc)
     }
     else
     {
-        char fname[OSFNMAX];
-        
-        /* 
-         *   get the filename string (converted to the file system
-         *   character set) 
-         */
-        G_stk->push(&fname_val);
-        pop_str_val_fname(vmg_ fname, sizeof(fname));
+        /* set up the recursive call context */
+        vm_rcdesc rc(vmg_ "setLogFile", bif_table, 1,
+                     G_stk->get(0), argc);
 
         /* open the appropriate log file */
         switch(log_type)
         {
         case LOG_SCRIPT:
-            G_console->open_log_file(fname);
+            ok = !G_console->open_log_file(vmg_ filespec, &rc);
             break;
 
         case LOG_CMD:
         case LOG_EVENT:
-            G_console->open_command_log(fname, log_type == LOG_EVENT);
+            ok = !G_console->open_command_log(
+                vmg_ filespec, &rc, log_type == LOG_EVENT);
             break;
 
         default:
             err_throw(VMERR_BAD_VAL_BIF);
         }
     }
+
+    /* discard arguments */
+    G_stk->discard(argc);
+
+    /* return true on success, nil on error */
+    retval_bool(vmg_ ok);
 }
 
 /* ------------------------------------------------------------------------ */
@@ -228,6 +236,10 @@ void CVmBifTIO::more(VMG_ uint argc)
     /* check arguments */
     check_argc(vmg_ argc, 0);
 
+    /* ignore in web host mode */
+    if (G_net_config != 0)
+        return;
+
     /* 
      *   if we're reading from a script, ignore this request - these types of
      *   interactive pauses are irrelevant when reading a script, since we're
@@ -249,12 +261,11 @@ void CVmBifTIO::more(VMG_ uint argc)
  */
 void CVmBifTIO::input(VMG_ uint argc)
 {
-    char buf[256];
-    
     /* check arguments */
     check_argc(vmg_ argc, 0);
 
     /* read a line of text from the keyboard */
+    char buf[256];
     if (G_console->read_line(vmg_ buf, sizeof(buf)))
     {
         /* end of file - return nil */
@@ -300,6 +311,13 @@ void CVmBifTIO::inputkey(VMG_ uint argc)
     /* flush any output */
     G_console->flush_all(vmg_ VM_NL_INPUT);
     
+    /* there's no console in web host mode - simply return eof */
+    if (G_net_config != 0)
+    {
+        retval_ui_str(vmg_ "[eof]");
+        return;
+    }
+
     /* get a keystroke */
     c[0] = (char)os_getc_raw();
     len = 1;
@@ -416,6 +434,11 @@ void CVmBifTIO::inputevent(VMG_ uint argc)
             break;
         }
     }
+    else if (G_net_config != 0)
+    {
+        /* there's no console in web host mode - return eof */
+        evt = OS_EVT_EOF;
+    }
     else
     {
         /* flush any buffered output */
@@ -461,6 +484,7 @@ void CVmBifTIO::inputevent(VMG_ uint argc)
     /* create the return list */
     lst_obj = CVmObjList::create(vmg_ FALSE, ele_count);
     lst = (CVmObjList *)vm_objp(vmg_ lst_obj);
+    lst->cons_clear();
 
     /* save the list on the stack to protect against garbage collection */
     val.set_obj(lst_obj);
@@ -631,7 +655,8 @@ int CVmBifTIO::map_ext_key(VMG_ char *namebuf, int extc)
         "[word-right]",                              /* CMD_WORD_RIGHT - 30 */
         "[del-word]",                                  /* CMD_WORDKILL - 31 */
         "[eof]",                                            /* CMD_EOF - 32 */
-        "[break]"                                         /* CMD_BREAK - 33 */
+        "[break]",                                        /* CMD_BREAK - 33 */
+        "[insert]"                                          /* CMD_INS - 34 */
     };
 
     /* if it's in the key name array, use the array entry */
@@ -750,9 +775,9 @@ void CVmBifTIO::inputdialog(VMG_ uint argc)
     char label_buf[256];
     vm_val_t label_val[10];
     const char *labels[10];
+    int lst_cnt;
     int std_btns;
     int btn_cnt;
-    const char *listp;
     char *dst;
     size_t dstrem;
     int default_resp;
@@ -778,47 +803,40 @@ void CVmBifTIO::inputdialog(VMG_ uint argc)
         /* get the standard button set ID */
         std_btns = pop_int_val(vmg0_);
     }
-    else
+    else if (G_stk->get(0)->is_listlike(vmg0_)
+             && (lst_cnt = G_stk->get(0)->ll_length(vmg0_)) >= 0)
     {
-        size_t i;
-        size_t cnt;
+        int i;
         vm_val_t *valp;
         
         /* we're not using any standard button set */
         std_btns = 0;
 
-        /* get the button label list */
-        listp = pop_list_val(vmg0_);
-
         /* 
          *   run through the list and get the button items into our array
          *   (we do this rather than traversing the list directly so that
          *   we don't have to worry about a constant list's data being
          *   paged out) 
          */
-        cnt = vmb_get_len(listp);
+        vm_val_t *lst = G_stk->get(0);
 
         /* limit the number of elements to our private array size */
-        if (cnt > sizeof(label_val)/sizeof(label_val[0]))
-            cnt = sizeof(label_val)/sizeof(label_val[0]);
-
-        /* skip the list length prefix */
-        listp += VMB_LEN;
+        if (lst_cnt > (int)sizeof(label_val)/sizeof(label_val[0]))
+            lst_cnt = sizeof(label_val)/sizeof(label_val[0]);
 
         /* copy the list */
-        for (i = cnt, valp = label_val ; i > 0 ;
-             --i, listp += VMB_DATAHOLDER, ++valp)
-        {
-            /* get this element into our array */
-            vmb_get_dh(listp, valp);
-        }
+        for (i = 1, valp = label_val ; i <= lst_cnt ; ++i, ++valp)
+            lst->ll_index(vmg_ valp, i);
+
+        /* done with the list - discard it */
+        G_stk->discard();
 
         /* set up to write into our label buffer */
         dst = label_buf;
         dstrem = sizeof(label_buf);
 
         /* now build our internal button list from the array elements */
-        for (i = 0, valp = label_val ; i < cnt ; ++i, ++valp)
+        for (i = 0, valp = label_val ; i < lst_cnt ; ++i, ++valp)
         {
             const char *p;
 
@@ -904,6 +922,11 @@ void CVmBifTIO::inputdialog(VMG_ uint argc)
             }
         }
     }
+    else
+    {
+        /* invalid button type */
+        err_throw(VMERR_BAD_TYPE_BIF);
+    }
 
     /* get the default response */
     if (G_stk->get(0)->typ == VM_NIL)
@@ -949,6 +972,11 @@ void CVmBifTIO::inputdialog(VMG_ uint argc)
         G_console->log_event(vmg_ VMCON_EVT_DIALOG,
                              numbuf, strlen(numbuf), FALSE);
     }
+    else if (G_net_config != 0)
+    {
+        /* there's no console in web host mode - return eof */
+        resp = 0;
+    }
     else
     {
         /* flush output before showing the dialog */
@@ -978,13 +1006,14 @@ void CVmBifTIO::askfile(VMG_ uint argc)
     char prompt[256];
     int dialog_type;
     os_filetype_t file_type;
-    long flags;
     int result;
     char fname[OSFNMAX*3 + 1];
     vm_obj_id_t lst_obj;
     CVmObjList *lst;
     vm_val_t val;
     int from_script = FALSE;
+    char warning[OSFNMAX + 255] = "";
+    int from_ui = FALSE;
     
     /* check arguments */
     check_argc(vmg_ argc, 4);
@@ -996,18 +1025,21 @@ void CVmBifTIO::askfile(VMG_ uint argc)
     dialog_type = pop_int_val(vmg0_);
     file_type = (os_filetype_t)pop_int_val(vmg0_);
 
-    /* pop the flags */
-    flags = pop_long_val(vmg0_);
+    /* 
+     *   Pop and discard the flags.  (This argument isn't used currently;
+     *   it's just there in case we need some option flags in the future.
+     *   Pop it as an integer to ensure the caller specified the correct
+     *   type, but discard the value.)
+     */
+    (void)pop_long_val(vmg0_);
 
     /* check for a script response */
     static int filter[] = { VMCON_EVT_FILE };
     int evt;
     unsigned long attrs;
     if (G_console->read_event_script(
-        vmg_ &evt, fname, sizeof(fname),
-        filter, sizeof(filter)/sizeof(filter[0]), &attrs))
+        vmg_ &evt, fname, sizeof(fname), filter, countof(filter), &attrs))
     {
-        char prompt[OSFNMAX + 255];
         int ok = TRUE;
         
         /* we got a response from the script */
@@ -1015,29 +1047,30 @@ void CVmBifTIO::askfile(VMG_ uint argc)
         result = (fname[0] != '\0' ? OS_AFE_SUCCESS : OS_AFE_CANCEL);
 
         /* 
-         *   If this is a "save" prompt, and the OVERWRITE flag isn't
-         *   provided, and the file already exists, show an interactive
-         *   warning that we're about to ovewrite the file.  
-         */
-        if (dialog_type == OS_AFP_SAVE
-            && result == OS_AFE_SUCCESS
-            && (attrs & VMCON_EVTATTR_OVERWRITE) == 0
-            && !osfacc(fname))
-        {
-            /* the file exists - warn about the overwrite */
-            ok = FALSE;
-            sprintf(prompt, "The script might overwrite the file %s. ", fname);
-        }
-
-        /* 
-         *   If this is a "save" prompt, AND the file doesn't already exist,
-         *   try creating the file to see if we can - this will test to see
-         *   if the target directory exists and is writable, for instance.  
+         *   If this is a "save" prompt, and the OVERWRITE flag isn't set,
+         *   and the file already exists, show an interactive warning that
+         *   we're about to ovewrite the file.  
          */
         if (ok
             && dialog_type == OS_AFP_SAVE
             && result == OS_AFE_SUCCESS
-            && osfacc(fname))
+            && (attrs & VMCON_EVTATTR_OVERWRITE) == 0
+            && CVmNetFile::exists(vmg_ fname, 0))
+        {
+            /* the file exists - warn about the overwrite */
+            ok = FALSE;
+            t3sprintf(warning, sizeof(warning),
+                      "OV The script might overwrite the file %s. ", fname);
+        }
+
+        /* 
+         *   If this is a "save" prompt, check to see if we can write the
+         *   file, and warn if not.
+         */
+        if (ok
+            && dialog_type == OS_AFP_SAVE
+            && result == OS_AFE_SUCCESS
+            && !CVmNetFile::can_write(vmg_ fname, 0))
         {
             /* try creating the file */
             osfildef *fp = osfopwb(fname, file_type);
@@ -1053,43 +1086,55 @@ void CVmBifTIO::askfile(VMG_ uint argc)
             {
                 /* didn't work - warn about it */
                 ok = FALSE;
-                sprintf(prompt, "The script is attempting to create file %s, "
-                        "but that file cannot be created.", fname);
+                t3sprintf(warning, sizeof(warning),
+                          "WR The script is attempting to write file %s, "
+                          "but that file cannot be written.", fname);
             }
         }
 
         /*
          *   If this is an "open" prompt, and the file isn't readable, warn
-         *   about it. 
+         *   about it.  
          */
         if (ok
             && dialog_type == OS_AFP_OPEN
             && result == OS_AFE_SUCCESS
-            && osfacc(fname))
+            && !CVmNetFile::exists(vmg_ fname, 0))
         {
             ok = FALSE;
-            sprintf(prompt, "The script is attempting to open file %s, "
-                    "but this file doesn't exist or isn't readable.", fname);
+            t3sprintf(warning, sizeof(warning),
+                      "RD The script is attempting to open file %s, "
+                      "but this file doesn't exist or isn't readable.",
+                      fname);
         }
 
-        /* check to see if we're warning about anything */
-        if (!ok)
+        /* 
+         *   If we generated a warning, and we're not in Web UI mode, display
+         *   the warning as a Yes/No/Cancel console input dialog.  We can't
+         *   do this in Web UI mode, since we can't use the regular console
+         *   in Web mode.  Instead, we'll return the warning information to
+         *   the caller for display.  
+         */
+        if (!ok && G_net_config == 0)
         {
             char fullprompt[OSFNMAX + 255 + 150];
 
             /* build the full prompt */
-            sprintf(fullprompt, "%s Do you wish to proceed? Select Yes to "
-                    "proceed with this file, No to choose a different file, "
-                    "or Cancel to stop script playback.", prompt);
+            t3sprintf(fullprompt, sizeof(fullprompt),
+                      "%s Do you wish to proceed? Select Yes to "
+                      "proceed with this file, No to choose a different "
+                      "file, or Cancel to stop replaying this script.",
+                      warning + 3);
 
             /* 
              *   display a dialog - note that this goes directly to user,
              *   bypassing the active script, since this is a question about
              *   how to handle a problem in the script 
              */
+        show_warning:
             switch (G_console->input_dialog(
                 vmg_ OS_INDLG_ICON_WARNING, fullprompt,
-                OS_INDLG_YESNOCANCEL, 0, 0, 2, 3))
+                OS_INDLG_YESNOCANCEL, 0, 0, 2, 3, TRUE))
             {
             case 1:
                 /* yes - proceed with fname */
@@ -1099,13 +1144,23 @@ void CVmBifTIO::askfile(VMG_ uint argc)
                 /* no - ask for a new file */
                 result = G_console->askfile(
                     vmg_ prompt, strlen(prompt),
-                    fname, sizeof(fname), dialog_type, file_type);
+                    fname, sizeof(fname), dialog_type, file_type, TRUE);
 
                 /* this didn't come from the script after all */
                 from_script = FALSE;
 
-                /* if they canceled, cancel the whole script playback */
-                if (result != OS_AFE_SUCCESS)
+                /* 
+                 *   If they canceled the file selection, go back to the
+                 *   warning dialog; if the dialog itself failed, cancel the
+                 *   script entirely.  If they selected a file, proceed with
+                 *   their new file replacing the script input.  Note that we
+                 *   don't have to repeat the tests above on the new file,
+                 *   since the user explicitly entered it and thus presumably
+                 *   knows that it's the one they really want to use.  
+                 */
+                if (result == OS_AFE_CANCEL)
+                    goto show_warning;
+                else if (result == OS_AFE_FAILURE)
                     close_script_file(vmg0_);
 
                 /* handled */
@@ -1119,26 +1174,60 @@ void CVmBifTIO::askfile(VMG_ uint argc)
                 result = OS_AFE_CANCEL;
                 break;
             }
+
+            /* we've displayed the warning, so don't return it */
+            warning[0] = '\0';
+        }
+    }
+    else if (G_net_config != 0)
+    {
+#ifdef TADSNET
+        /*
+         *   If we're running in the local stand-alone configuration, we have
+         *   a bit of a special case: we're accessing files on the local file
+         *   system, but we're presenting the rest of the program's UI
+         *   through a browser window, which might be in a separate child
+         *   process; we have to present the file dialog UI in the same
+         *   manner.  We have a special OS function for this situation.
+         *   
+         *   This only applies in the stand-alone configuration, where we
+         *   have no "hostname" parameter in the net config.  If we have a
+         *   hostname, it means that we're operating in full client/server
+         *   mode, in which case we go through the Web UI directly.  
+         */
+        if (G_net_config->get("hostname") == 0)
+        {
+            result = osnet_askfile(prompt, fname, sizeof(fname),
+                                   dialog_type, file_type);
+        }
+        else
+#endif
+        {
+            /* there's no console in web host mode - return eof */
+            result = OS_AFE_FAILURE;
         }
     }
     else
     {
-        /* ask for a file via the UI */
+        /* ask for a file via the console UI */
         result = G_console->askfile(
             vmg_ prompt, strlen(prompt),
             fname, sizeof(fname), dialog_type, file_type);
+
+        /* this file came from the console UI */
+        from_ui = TRUE;
     }
 
     /* 
-     *   Allocate a list to store the return value.  If we successfully
-     *   got a filename, we need a two-element list - one element for the
-     *   success code and another for the string with the filename.  If we
-     *   didn't succeed in getting the filename, we only need a single
-     *   element, which will contain the error code. 
+     *   Create the return list.  In all cases, we need one element for the
+     *   status code.  On success, we need three addition elements (the
+     *   filename string, the descriptive text [future expansion], and the
+     *   warning message text).  
      */
     lst_obj = CVmObjList::create(vmg_ FALSE,
-                                 result == OS_AFE_SUCCESS ? 2 : 1);
+                                 result == OS_AFE_SUCCESS ? 4 : 1);
     lst = (CVmObjList *)vm_objp(vmg_ lst_obj);
+    lst->cons_clear();
 
     /* save it on the stack as protection against garbage collection */
     val.set_obj(lst_obj);
@@ -1148,21 +1237,66 @@ void CVmBifTIO::askfile(VMG_ uint argc)
     val.set_int(result);
     lst->cons_set_element(0, &val);
 
-    /* if we got a filename, set the second element to the filename string */
+    /* add the extra elements for the success case */
     if (result == OS_AFE_SUCCESS)
     {
-        /* 
-         *   Create a string for the filename.  If it's coming from a script,
-         *   we need to translate it to the local character set; otherwise
-         *   it's already in UTF-8. 
-         */
-        val.set_obj(
-            from_script
-            ? str_from_ui_str(vmg_ fname)
-            : CVmObjString::create(vmg_ FALSE, fname, strlen(fname)));
+        char *fnamep = fname, *fname2 = 0;
+        err_try
+        {
+            /* if the name came from a script, map from the local char set */
+            if (from_script)
+            {
+                G_cmap_from_ui->map_str_alo(&fname2, fname);
+                fnamep = fname2;
+            }
+
+            /* create a FileName object to represent the filename */
+            val.set_obj(CVmObjFileName::create_from_local(
+                vmg_ fnamep, strlen(fnamep)));
+
+            /* 
+             *   if the name came from the console UI, flag the FileName as
+             *   being user-selected, which allows it to override the file
+             *   safety settings; since the user manually selected the name,
+             *   they implicitly granted permission to use the file for the
+             *   type of operation proposed by the dialog
+             */
+            if (from_ui)
+            {
+                vm_objid_cast(CVmObjFileName, val.val.obj)->set_from_ui(
+                    dialog_type);
+            }
+        }
+        err_finally
+        {
+            lib_free_str(fname2);
+        }
+        err_end;
 
         /* store the string as the second list element */
         lst->cons_set_element(1, &val);
+
+        /* 
+         *   Add nil for the descriptive text, since we don't currently
+         *   implement this feature.  (This is a future enhancement that some
+         *   system file selector dialogs might provide, allowing the user to
+         *   enter descriptive text in addition to the filename.  The Web UI
+         *   storage server already uses this to save user descriptive text
+         *   with saved game files.)  
+         */
+        val.set_nil();
+        lst->cons_set_element(2, &val);
+
+        /* 
+         *   Add the script warning message, or nil if there's no warning.
+         *   The warning message contains the filename text from the script,
+         *   so translate it to the local character set.  
+         */
+        if (warning[0] != '\0')
+            val.set_obj(str_from_ui_str(vmg_ warning));
+        else
+            val.set_nil();
+        lst->cons_set_element(3, &val);
     }
 
     /* log the event */
@@ -1408,9 +1542,6 @@ void CVmBifTIO::res_exists(VMG_ uint argc)
  */
 void CVmBifTIO::set_script_file(VMG_ uint argc)
 {
-    char fname[OSFNMAX];
-    int flags;
-
     /* check arguments */
     check_argc_range(vmg_ argc, 1, 2);
 
@@ -1430,11 +1561,14 @@ void CVmBifTIO::set_script_file(VMG_ uint argc)
 
         /* close the script file */
         close_script_file(vmg0_);
+
+        /* this form always succeeds */
+        retval_true(vmg0_);
     }
     else if (G_stk->get(0)->typ == VM_INT)
     {
         /* get the request code */
-        flags = pop_int_val(vmg0_);
+        int flags = pop_int_val(vmg0_);
         
         /* any additional argument is superfluous in this case */
         if (argc >= 2)
@@ -1462,25 +1596,35 @@ void CVmBifTIO::set_script_file(VMG_ uint argc)
                 retval_nil(vmg0_);
             }
             break;
+
+        default:
+            /* other flags are invalid */
+            err_throw(VMERR_BAD_VAL_BIF);
         }
     }
     else
     {
-        /* 
-         *   get the filename string (converted to the file system
-         *   character set) 
-         */
-        pop_str_val_fname(vmg_ fname, sizeof(fname));
+        /* get the filename argument */
+        const vm_val_t *filespec = G_stk->get(0);
     
-        /* if they provided flags, pop the flags */
-        flags = 0;
-        if (argc >= 2)
-            flags = pop_int_val(vmg0_);
+        /* if they provided flags, get the flags */
+        int flags = (argc >= 2 ? G_stk->get(1)->num_to_int(vmg0_) : 0);
+
+        /* set up our recursive call descriptor */
+        vm_rcdesc rc(vmg_ "setScriptFile", bif_table, 15,
+                     G_stk->get(0), argc);
 
         /* open the script file */
-        G_console->open_script_file(fname,
-                                    (flags & VMBIFTADS_SCRIPT_QUIET) != 0,
-                                    !(flags & VMBIFTADS_SCRIPT_NONSTOP));
+        int ok = !G_console->open_script_file(
+            vmg_ filespec, &rc,
+            (flags & VMBIFTADS_SCRIPT_QUIET) != 0,
+            !(flags & VMBIFTADS_SCRIPT_NONSTOP));
+
+        /* discard arguments */
+        G_stk->discard(argc);
+
+        /* return true on success, nil on failure */
+        retval_bool(vmg_ ok);
     }
 }
 
@@ -1492,7 +1636,7 @@ void CVmBifTIO::close_script_file(VMG0_)
     int old_more_mode;
 
     /* close the script */
-    old_more_mode = G_console->close_script_file();
+    old_more_mode = G_console->close_script_file(vmg0_);
 
     /* restore the MORE mode in effect when the script was opened */
     G_console->set_more_state(old_more_mode);
@@ -1651,6 +1795,7 @@ void CVmBifTIO::input_timeout(VMG_ uint argc)
     /* create the return list */
     lst_obj = CVmObjList::create(vmg_ FALSE, ele_count);
     lst = (CVmObjList *)vm_objp(vmg_ lst_obj);
+    lst->cons_clear();
 
     /* save the list on the stack to protect against garbage collection */
     val.set_obj(lst_obj);
@@ -1793,6 +1938,13 @@ void CVmBifTIO::banner_create(VMG_ uint argc)
     /* retrieve the flags */
     style = pop_long_val(vmg0_);
 
+    if (G_net_config != 0)
+    {
+        /* there's no console in web host mode - return failure */
+        retval_nil(vmg0_);
+        return;
+    }
+
     /* try creating the banner */
     hdl = G_console->get_banner_manager()->create_banner(
         vmg_ parent_id, where, other_id, wintype,
@@ -1817,7 +1969,7 @@ void CVmBifTIO::banner_delete(VMG_ uint argc)
     check_argc(vmg_ argc, 1);
 
     /* delete the banner */
-    G_console->get_banner_manager()->delete_banner(pop_int_val(vmg0_));
+    G_console->get_banner_manager()->delete_banner(vmg_ pop_int_val(vmg0_));
 }
 
 /*
@@ -2134,100 +2286,95 @@ void CVmBifTIO::banner_get_info(VMG_ uint argc)
  */
 void CVmBifTIO::log_console_create(VMG_ uint argc)
 {
-    char fname[OSFNMAX];
-    osfildef *fp;
-    CCharmapToLocal *cmap;
-    int width;
-    int hdl;
-    
     /* check arguments */
     check_argc(vmg_ argc, 3);
 
-    /* retrieve the log file name */
-    pop_str_val_fname(vmg_ fname, sizeof(fname));
+    /* get the arguments: log file name, character set, and width */
+    const vm_val_t *filespec = G_stk->get(0);
+    const vm_val_t *charset = G_stk->get(1);
+    int width = G_stk->get(2)->num_to_int(vmg0_);
 
     /* 
      *   Retrieve the character mapper, which can be given as either a
      *   CharacterSet object or a string giving the character set name.  
      */
-    if (G_stk->get(0)->typ == VM_NIL)
+    CCharmapToLocal *cmap = 0;
+    if (charset->typ == VM_NIL)
     {
         /* nil - use the default log file character set */
         cmap = G_cmap_to_log;
-
-        /* add our reference to it */
         cmap->add_ref();
     }
-    else if (G_stk->get(0)->typ == VM_OBJ
-             && CVmObjCharSet::is_charset(vmg_ G_stk->get(0)->val.obj))
+    else if (charset->typ == VM_OBJ
+             && CVmObjCharSet::is_charset(vmg_ charset->val.obj))
     {
-        vm_obj_id_t obj;
-        
         /* it's a CharacterSet object - pop the reference */
-        obj = CVmBif::pop_obj_val(vmg0_);
-
+        vm_obj_id_t obj = charset->val.obj;
+        
         /* retrieve the character mapper from the character set */
         cmap = ((CVmObjCharSet *)vm_objp(vmg_ obj))->get_to_local(vmg0_);
-
-        /* add our reference to it */
         cmap->add_ref();
     }
     else
     {
-        const char *str;
-        size_t len;
-        char *nm;
-
         /* it's not a CharacterSet, so it must be a character set name */
-        str = G_stk->get(0)->get_as_string(vmg0_);
+        const char *str = charset->get_as_string(vmg0_);
         if (str == 0)
             err_throw(VMERR_BAD_TYPE_BIF);
         
         /* get the length and skip the length prefix */
-        len = vmb_get_len(str);
+        size_t len = vmb_get_len(str);
         str += VMB_LEN;
 
         /* get a null-terminated version of the name */
-        nm = lib_copy_str(str, len);
+        char *nm = lib_copy_str(str, len);
 
         /* 
          *   Create a character mapping for the given name.  Note that this
          *   will automatically add a reference to the mapper on our behalf,
          *   so we don't have to add our own extra reference. 
          */
-        cmap = CCharmapToLocal::load(G_host_ifc->get_cmap_res_loader(), nm);
+        cmap = CCharmapToLocal::load(G_host_ifc->get_sys_res_loader(), nm);
 
         /* done with the null-terminated version of the name string */
         lib_free_str(nm);
-
-        /* discard the argument */
-        G_stk->discard();
     }
 
     /* if we didn't get a character map, use us-ascii by default */
     if (cmap == 0)
-        cmap = CCharmapToLocal::load(G_host_ifc->get_cmap_res_loader(),
+        cmap = CCharmapToLocal::load(G_host_ifc->get_sys_res_loader(),
                                      "us-ascii");
 
+    /* open the file and start logging */
+    osfildef *fp = 0;
+    CVmNetFile *nf = 0;
+    int hdl = 0;
     err_try
     {
+        /* create the network file descriptor */
+        vm_rcdesc rc(vmg_ "logConsoleCreate", bif_table, 30,
+                     G_stk->get(0), argc);
+        nf = CVmNetFile::open(vmg_ filespec, &rc,
+                              NETF_NEW, OSFTLOG, "text/plain");
+
         /* make sure the file safety level allows the operation */
-        CVmObjFile::check_safety_for_open(vmg_ fname, VMOBJFILE_ACCESS_WRITE);
+        CVmObjFile::check_safety_for_open(vmg_ nf, VMOBJFILE_ACCESS_WRITE);
         
         /* open the file for writing (in text mode) */
-        fp = osfopwt(fname, OSFTLOG);
+        fp = osfopwt(nf->lclfname, OSFTLOG);
     
         /* if that failed, we can't contine */
         if (fp == 0)
             G_interpreter->throw_new_class(vmg_ G_predef->file_creation_exc,
                                            0, "error creating log file");
 
-        /* retrieve the width */
-        width = pop_int_val(vmg0_);
-
         /* create the log console */
         hdl = G_console->get_log_console_manager()->create_log_console(
-            fname, fp, cmap, width);
+            vmg_ nf, fp, cmap, width);
+
+        /* we've handed off the file information to the console */
+        nf = 0;
+        fp = 0;
     }
     err_finally
     {
@@ -2237,9 +2384,20 @@ void CVmBifTIO::log_console_create(VMG_ uint argc)
          *   by now 
          */
         cmap->release_ref();
+
+        /* close the file if we opened it and didn't hand it off */
+        if (fp != 0)
+            osfcls(fp);
+
+        /* if we have a network file object, abandon it */
+        if (nf != 0)
+            nf->abandon(vmg0_);
     }
     err_end;
 
+    /* discard arguments */
+    G_stk->discard(argc);
+
     /* 
      *   If we succeeded, return the handle; otherwise, return nil.  A handle
      *   of zero indicates failure.  
@@ -2273,7 +2431,10 @@ void CVmBifTIO::log_console_close(VMG_ uint argc)
     console->flush(vmg_ VM_NL_NONE);
 
     /* delete the console */
-    G_console->get_log_console_manager()->delete_log_console(handle);
+    G_console->get_log_console_manager()->delete_log_console(vmg_ handle);
+
+    /* no return value */
+    retval_nil(vmg0_);
 }
 
 /*
@@ -2283,7 +2444,6 @@ void CVmBifTIO::log_console_say(VMG_ uint argc)
 {
     int hdl;
     CVmConsole *console;
-    CVmConsoleMainLog main_log_console;
 
     /* check arguments */
     check_argc_range(vmg_ argc, 1, 32767);
@@ -2299,7 +2459,7 @@ void CVmBifTIO::log_console_say(VMG_ uint argc)
     if (hdl == -1)
     {
         /* use the main log */
-        console = &main_log_console;
+        console = new CVmConsoleMainLog();
     }
     else
     {
@@ -2314,6 +2474,96 @@ void CVmBifTIO::log_console_say(VMG_ uint argc)
      *   which we've already retrieved, is the console handle, so don't count
      *   it among the arguments to display) 
      */
-    say_to_console(vmg_ console, argc - 1);
+    err_try
+    {
+        say_to_console(vmg_ console, argc - 1);
+    }
+    err_finally
+    {
+        /* if we created a "main console" object, we're done with it */
+        if (hdl == -1)
+            console->delete_obj(vmg0_);
+    }
+    err_end;
 }
 
+/* ------------------------------------------------------------------------ */
+/*
+ *   Log an input event obtained from an external source, such as the Web UI 
+ */
+void CVmBifTIO::log_input_event(VMG_ uint argc)
+{
+    /* check arguments */
+    check_argc(vmg_ argc, 1);
+
+    /* retrieve the list */
+    const vm_val_t *lst = G_stk->get(0);
+    if (!lst->is_listlike(vmg0_))
+        err_throw(VMERR_BAD_TYPE_BIF);
+
+    /* make sure we have at least one element */
+    int len = lst->ll_length(vmg0_);
+    if (len < 1)
+        err_throw(VMERR_BAD_VAL_BIF);
+
+    /* retrieve the event type code or tag name from the first element */
+    vm_val_t ele1;
+    lst->ll_index(vmg_ &ele1, 1);
+
+    /* if there's a second element, retrieve the parameter string */
+    const char *param = 0;
+    size_t paramlen = 0;
+    if (len >= 2)
+    {
+        /* there's a parameter - retrieve it */
+        vm_val_t ele2;
+        lst->ll_index(vmg_ &ele2, 2);
+
+        /* 
+         *   Get its string value.  If it's a filename object, use the
+         *   filename path string. 
+         */
+        CVmObjFileName *fname_param = vm_val_cast(CVmObjFileName, &ele2);
+        if (fname_param != 0)
+            param = fname_param->get_path_string();
+        else
+            param = ele2.get_as_string(vmg0_);
+
+        /* if we didn't get a parameter string value, it's an error */
+        if (param == 0)
+            err_throw(VMERR_BAD_VAL_BIF);
+
+        /* get the parameter length and buffer pointer */
+        paramlen = vmb_get_len(param);
+        param += VMB_LEN;
+    }
+
+    /* log the event, according to the type of the event code */
+    if (ele1.is_numeric(vmg0_))
+    {
+        /* it's an integer event type code - log it */
+        G_console->log_event(
+            vmg_ ele1.num_to_int(vmg0_), param, paramlen, TRUE);
+    }
+    else if (ele1.get_as_string(vmg0_) != 0)
+    {
+        /* it's a string event tag - retrieve it in the UI character set */
+        char tag[128];
+        G_stk->push(&ele1);
+        pop_str_val_ui(vmg_ tag, sizeof(tag));
+
+        /* log the event with the given tag */
+        G_console->log_event(vmg_ tag, param, paramlen, TRUE);
+    }
+    else
+    {
+        /* other types are invalid */
+        err_throw(VMERR_BAD_VAL_BIF);
+    }
+
+    /* discard arguments */
+    G_stk->discard(1);
+
+    /* no return value */
+    retval_nil(vmg0_);
+}
