@@ -280,6 +280,7 @@ public:
      */
     void run(VMG_ const char *const *argv, int argc,
              class CVmRuntimeSymbols *global_symtab,
+             class CVmRuntimeSymbols *macro_symtab,
              const char *saved_state);
 
     /* run all static initializers */
@@ -293,11 +294,14 @@ public:
     void unload(VMG0_);
 
     /* 
-     *   Create a global LookupTable to hold hte symbols in the global
+     *   Create a global LookupTable to hold the symbols in the global
      *   symbol table.
      */
     void create_global_symtab_lookup_table(VMG0_);
 
+    /* create a global LookupTable to hold the symbols in the macro table */
+    void create_macro_symtab_lookup_table(VMG0_);
+
     /* determine if the given block type identifiers match */
     static int block_type_is(const char *type1, const char *type2)
     {
@@ -318,6 +322,9 @@ public:
     /* get the filename of the loaded image */
     const char *get_filename() const { return fname_; }
 
+    /* get the fully-qualified, absolute directory path of the loaded image */
+    const char *get_path() const { return path_; }
+
     /* 
      *   check to see if the image file has a global symbol table (GSYM
      *   block) 
@@ -330,6 +337,9 @@ public:
      */
     vm_obj_id_t get_reflection_symtab() const { return reflection_symtab_; }
 
+    /* get the object ID of the LookupTable with the macro table */
+    vm_obj_id_t get_reflection_macros() const { return reflection_macros_; }
+
     /*
      *   perform dynamic linking after loading, resetting, or restoring 
      */
@@ -344,6 +354,9 @@ public:
     /* allocate a new property ID */
     vm_prop_id_t alloc_new_prop(VMG0_);
 
+    /* get the last property ID currently in use */
+    vm_prop_id_t get_last_prop(VMG0_);
+
     /* save/restore the synthesized export table */
     void save_synth_exports(VMG_ class CVmFile *fp);
     int restore_synth_exports(VMG_ class CVmFile *fp,
@@ -353,7 +366,7 @@ public:
     ulong get_static_cs_ofs() const { return static_cs_ofs_; }
 
     /* get the entrypoint function's code pool offset */
-    uint32 get_entrypt() const { return entrypt_; }
+    uint32_t get_entrypt() const { return entrypt_; }
 
 private:
     /* load external resource files associated with an image file */
@@ -398,6 +411,9 @@ private:
     /* load a Global Symbols (GSYM) block into the runtime symbol table */
     void load_runtime_symtab_from_gsym(VMG_ ulong siz);
 
+    /* load a Macro Symbols (MACR) block into the runtime symbol table */
+    void load_runtime_symtab_from_macr(VMG_ ulong siz);
+
     /* load a Macro Symbols (MACR) block */
     void load_macros(VMG_ ulong siz);
 
@@ -457,6 +473,12 @@ private:
     /* image filename */
     char *fname_;
 
+    /* 
+     *   fully-qualified, absolute directory path to the file (this is just
+     *   the directory path, sans the filename) 
+     */
+    char *path_;
+
     /* the base seek offset of the image stream within the image file */
     long base_seek_ofs_;
 
@@ -467,7 +489,7 @@ private:
     char timestamp_[24];
 
     /* code pool offset of entrypoint function */
-    uint32 entrypt_;
+    uint32_t entrypt_;
 
     /* pool tracking objects */
     class CVmImagePool *pools_[2];
@@ -500,9 +522,19 @@ private:
      */
     class CVmRuntimeSymbols *runtime_symtab_;
 
+    /* 
+     *   The runtime macro definitions table, if we have one.  As with the
+     *   runtime global symbol table, we build this from the debug records,
+     *   or from the records passed in from the compiler during preinit. 
+     */
+    class CVmRuntimeSymbols *runtime_macros_;
+
     /* object ID of LookupTable containing the global symbol table */
     vm_obj_id_t reflection_symtab_;
 
+    /* object ID of LookupTable containing the macro symbol table */
+    vm_obj_id_t reflection_macros_;
+
     /* head/tail of list of static initializer pages */
     class CVmStaticInitPage *static_head_;
     class CVmStaticInitPage *static_tail_;
@@ -589,7 +621,7 @@ public:
     virtual ~CVmImageLoaderMres() { }
 
     /* load a resource */
-    virtual void add_resource(uint32 seek_ofs, uint32 siz,
+    virtual void add_resource(uint32_t seek_ofs, uint32_t siz,
                               const char *res_name, size_t res_name_len) = 0;
 
     /* load a resource link */
@@ -609,6 +641,9 @@ class CVmImageFile
 public:
     /* delete the loader */
     virtual ~CVmImageFile() { }
+
+    /* duplicate the image file interface, a la stdio freopen() */
+    virtual CVmImageFile *dup(const char *mode) = 0;
     
     /* 
      *   Copy data from the image file to the caller's buffer.  Reads from
@@ -669,8 +704,23 @@ public:
         len_ = len;
     }
 
+    CVmStream *clone(VMG_ const char *mode)
+    {
+        /* duplicate our file handle */
+        CVmImageFile *fpdup = fp_->dup(mode);
+        if (fpdup == 0)
+            return 0;
+
+        /* create a new image file stream wrapper for the duplicate handle */
+        return new CVmImageFileStream(fpdup, len_);
+    }
+
     /* read bytes into a buffer */
     virtual void read_bytes(char *buf, size_t len);
+    virtual size_t read_nbytes(char *buf, size_t len);
+
+    /* read a line (not used for this object) */
+    virtual char *read_line(char *buf, size_t len) { return 0; }
 
     /* write bytes */
     virtual void write_bytes(const char *, size_t);
@@ -679,6 +729,8 @@ public:
     virtual long get_seek_pos() const { return fp_->get_seek(); }
     virtual void set_seek_pos(long pos) { fp_->seek(pos); }
 
+    virtual long get_len() { return len_; }
+
 private:
     /* our underlying image file reader */
     CVmImageFile *fp_;
@@ -707,7 +759,19 @@ public:
         /* we don't have any suballocation blocks yet */
         mem_head_ = mem_tail_ = 0;
     }
-    
+
+    /* duplicate the file interface */
+    virtual CVmImageFile *dup(const char *mode)
+    {
+        /* duplicate our file handle */
+        CVmFile *fpdup = fp_->dup(mode);
+        if (fpdup == 0)
+            return 0;
+
+        /* return a new wrapper object for the duplicate handle */
+        return new CVmImageFileExt(fpdup);
+    }
+
     /* 
      *   CVmImageFile interface implementation 
      */
@@ -824,6 +888,12 @@ public:
         pos_ = 0;
     }
 
+    /* duplicate the file interface */
+    CVmImageFile *dup(const char *mode)
+    {
+        return new CVmImageFileMem(mem_, len_);
+    }
+
     /* 
      *   CVmImageFile interface implementation 
      */
