@@ -21,7 +21,7 @@ extern ContainerEntry *containers; /* Container table pointer */
 
 
 /* FUNCTIONS */
-extern int containerSize(int container, bool directly);
+extern int containerSize(int container, ATrans trans);
 extern bool passesContainerLimits(Aint container, Aint addedInstance);
 extern void describeContainer(int container);
 extern void list(int cnt);
