@@ -1,9 +1,9 @@
 /*----------------------------------------------------------------------*\
- 
+
   MAIN.C
- 
+
   Main module of interpreter for ALAN Adventure Language
- 
+
   \*----------------------------------------------------------------------*/
 #include "main.h"
 
@@ -62,21 +62,21 @@ VerbEntry *vrbs;		/* Verb table pointer */
 
 #ifdef CHECKOBJ
 /*======================================================================
- 
+
   checkobj()
- 
+
   Check that the object given is valid, else print an error message
   or find out what he wanted.
- 
+
   This routine is not used any longer, kept for sentimental reasons ;-)
- 
+
 */
 void checkobj(Aword *obj) {
     Aword oldobj;
-	
+
     if (*obj != EOF)
         return;
-	
+
     oldobj = EOF;
     for (cur.obj = OBJMIN; cur.obj <= OBJMAX; cur.obj++) {
         /* If an object is present and it is possible to perform his action */
@@ -86,10 +86,10 @@ void checkobj(Aword *obj) {
             else
                 error(WANT);          /* And we didn't find multiple objects */
     }
-	
+
     if (oldobj == EOF)
         error(WANT);              /* But we found ONE */
-	
+
     *obj = cur.obj = oldobj;
     output("($o)");             /* Then he surely meant this object */
 }
@@ -102,32 +102,36 @@ void checkobj(Aword *obj) {
  *
  * Event Handling
  *
- * eventchk()
- *
  *----------------------------------------------------------------------*/
 
+/*----------------------------------------------------------------------*/
+static char *eventName(int event) {
+    return stringAt(events[event].id);
+}
+
 
 /*----------------------------------------------------------------------*/
 static void runPendingEvents(void)
 {
     int i;
-	
+
+    resetRules();
     while (eventQueueTop != 0 && eventQueue[eventQueueTop-1].after == 0) {
         eventQueueTop--;
         if (isALocation(eventQueue[eventQueueTop].where))
             current.location = eventQueue[eventQueueTop].where;
         else
-            current.location = where(eventQueue[eventQueueTop].where, FALSE);
-        if (sectionTraceOption) {
-            printf("\n<EVENT %d (at ", eventQueue[eventQueueTop].event);
+            current.location = where(eventQueue[eventQueueTop].where, TRANSITIVE);
+        if (traceSectionOption) {
+            printf("\n<EVENT %s[%d] (at ", eventName(eventQueue[eventQueueTop].event),
+                   eventQueue[eventQueueTop].event);
             traceSay(current.location);
-            printf("):>\n");
+            printf(" [%d]):>\n", current.location);
         }
         interpret(events[eventQueue[eventQueueTop].event].code);
-        if (isPreBeta2(header->version))
-            evaluateRules();
+        evaluateRules(rules);
     }
-	
+
     for (i = 0; i<eventQueueTop; i++)
         eventQueue[i].after--;
 }
@@ -136,10 +140,10 @@ static void runPendingEvents(void)
 
 
 /*----------------------------------------------------------------------*\
- 
+
   Main program and initialisation
- 
-  \*----------------------------------------------------------------------*/
+
+\*----------------------------------------------------------------------*/
 
 
 static FILE *codfil;
@@ -147,6 +151,97 @@ static char codfnm[256] = "";
 static char txtfnm[256] = "";
 
 
+/*----------------------------------------------------------------------
+  Calculate where to start calculating the CRC. Is different for
+  different versions. CRC is calculated from pre-beta2 memory start to
+  be compatible.  If header size changes this should return beta2
+  header size for later versions.
+*/
+static int crcStart(char version[4]) {
+    /* Some earlier versions had a shorter header */
+    if (isPreAlpha5(version))
+        return sizeof(Pre3_0alpha5Header)/sizeof(Aword);
+    else if (isPreBeta2(version))
+        return sizeof(Pre3_0beta2Header)/sizeof(Aword);
+    else
+        return sizeof(ACodeHeader)/sizeof(Aword);
+}
+
+
+/*----------------------------------------------------------------------*/
+static void readTemporaryHeader(ACodeHeader *tmphdr) {
+    rewind(codfil);
+    fread(tmphdr, sizeof(*tmphdr), 1, codfil);
+    rewind(codfil);
+    if (strncmp((char *)tmphdr, "ALAN", 4) != 0)
+        playererr("Not an Alan game file, does not start with \"ALAN\"");
+}
+
+
+/*----------------------------------------------------------------------*/
+static void reverseMemory() {
+    if (littleEndian()) {
+        if (debugOption||traceSectionOption||traceInstructionOption)
+            output("<Hmm, this is a little-endian machine, fixing byte ordering....");
+        reverseACD();			/* Reverse content of the ACD file */
+        if (debugOption||traceSectionOption||traceInstructionOption)
+            output("OK.>$n");
+    }
+}
+
+
+/*----------------------------------------------------------------------*/
+static void setupHeader(ACodeHeader tmphdr) {
+    if (isPreBeta2(tmphdr.version)) {
+        header = duplicate(&memory[0], sizeof(ACodeHeader));
+        if (isPreAlpha5(tmphdr.version)) {
+            header->ifids = 0;
+        }
+        header->prompt = 0;
+    } else if (isPreBeta3(tmphdr.version)) {
+        header = (ACodeHeader *) pointerTo(0);
+    } else {
+        header = pointerTo(0);
+    }
+}
+
+
+/*----------------------------------------------------------------------*/
+static void loadAndCheckMemory(ACodeHeader tmphdr, Aword crc, char err[]) {
+    int i;
+    /* No memory allocated yet? */
+    if (memory == NULL) {
+        memory = allocate(tmphdr.size*sizeof(Aword));
+    }
+
+    memTop = fread(pointerTo(0), sizeof(Aword), tmphdr.size, codfil);
+    if (memTop != tmphdr.size)
+        syserr("Could not read all ACD code.");
+
+    /* Calculate checksum */
+    for (i = crcStart(tmphdr.version); i < memTop; i++) {
+        crc += memory[i]&0xff;
+        crc += (memory[i]>>8)&0xff;
+        crc += (memory[i]>>16)&0xff;
+        crc += (memory[i]>>24)&0xff;
+#ifdef CRCLOG
+        printf("%6x\t%6lx\t%6lx\n", i, crc, memory[i]);
+#endif
+    }
+    if (crc != tmphdr.acdcrc) {
+        sprintf(err, "Checksum error in Acode (.a3c) file (0x%lx instead of 0x%lx).",
+                (unsigned long) crc, (unsigned long) tmphdr.acdcrc);
+        if (!ignoreErrorOption)
+            syserr(err);
+        else {
+            output("<WARNING! $$");
+            output(err);
+            output("$$ Ignored, proceed at your own risk.>$n");
+        }
+    }
+}
+
+
 /*----------------------------------------------------------------------*/
 static char *decodeState(int c) {
     static char state[2] = "\0\0";
@@ -161,6 +256,16 @@ static char *decodeState(int c) {
     }
 }
 
+/*======================================================================*/
+char *decodedGameVersion(char version[]) {
+    static char str[100];
+    sprintf(str, "%d.%d%s%d",
+            (int)version[3],
+            (int)version[2],
+            decodeState(version[0]),
+            (int)version[1]);
+    return str;
+}
 
 /*----------------------------------------------------------------------*/
 static void incompatibleDevelopmentVersion(ACodeHeader *header) {
@@ -204,7 +309,7 @@ static void alphaRunningLaterGame(char gameState) {
 static void nonDevelopmentRunningDevelopmentStateGame(char version[]) {
     char errorMessage[200];
     char versionString[100];
-	
+
     strcpy(errorMessage, "Games generated by a development state compiler");
     sprintf(versionString, "(this game is v%d.%d.%d%s)", version[0], version[1],
             version[2], decodeState(version[3]));
@@ -214,8 +319,8 @@ static void nonDevelopmentRunningDevelopmentStateGame(char version[]) {
 }
 
 
-/*----------------------------------------------------------------------*/
-static void checkVersion(ACodeHeader *header)
+/*======================================================================*/
+void checkVersion(ACodeHeader *header)
 {
     /* Strategy for version matching is:
        1) Development interpreters/games require exact match
@@ -226,36 +331,38 @@ static void checkVersion(ACodeHeader *header)
        alpha interpreters, example is introduction of a new opcode if it is
        done at the end of the list)
        5) Release interpreters should run alpha and beta games without problems
+
+       NOTE that we are working with a non-reversed version string/word here.
     */
-	
+
     char interpreterVersion[4];
     bool developmentVersion;
     bool alphaVersion;
     int compareLength;
     char gameState = header->version[3];
-	
+
     /* Construct our own version */
     interpreterVersion[0] = alan.version.version;
     interpreterVersion[1] = alan.version.revision;
     interpreterVersion[2] = alan.version.correction;
     interpreterVersion[3] = alan.version.state[0];
-	
+
     /* Check version of .ACD file */
     if (debugOption && !regressionTestOption) {
-        printf("<Version of '%s' is %d.%d%s%d>",
+        printf("<Version of '%s' is %d.%d%s%d!>\n",
                adventureFileName,
-               (int)(header->version[0]),
-               (int)(header->version[1]),
-               decodeState(gameState),
-               (int)(header->version[2]));
+               (int)header->version[0],
+               (int)header->version[1],
+               decodeState(header->version[3]),
+               (int)header->version[2]);
         newline();
     }
-	
+
     /* Development version require exact match, else only 2 digit match */
     developmentVersion = (strcmp(alan.version.state, "dev") == 0);
     alphaVersion = (strcmp(alan.version.state, "alpha") == 0);
     compareLength = (developmentVersion? 3 : 2);
-	
+
     if (gameState == 'd' && !developmentVersion)
         /* Development state game requires development state interpreter... */
         nonDevelopmentRunningDevelopmentStateGame(header->version);
@@ -280,112 +387,24 @@ static void checkVersion(ACodeHeader *header)
     }
 }
 
-
-/*----------------------------------------------------------------------
-  Calculate where the actual memory starts. Might be different for
-  different versions.
-*/
-static int memoryStart(char version[4]) {
-    /* Pre 3.0alpha5 had a shorter header */
-    if (isPreAlpha5(version))
-        return sizeof(Pre3_0alpha5Header)/sizeof(Aword);
-    else if (isPreBeta2(version))
-        return sizeof(Pre3_0beta2Header)/sizeof(Aword);
-    else
-        return sizeof(ACodeHeader)/sizeof(Aword);
-}
-
-
-/*----------------------------------------------------------------------*/
-static void readTemporaryHeader(ACodeHeader *tmphdr) {
-    rewind(codfil);
-    fread(tmphdr, sizeof(*tmphdr), 1, codfil);
-    rewind(codfil);
-    if (strncmp((char *)tmphdr, "ALAN", 4) != 0)
-        playererr("Not an Alan game file, does not start with \"ALAN\"");
-}
-
-
-/*----------------------------------------------------------------------*/
-static void reverseMemory() {
-    if (littleEndian()) {
-        if (debugOption||sectionTraceOption||singleStepOption)
-            output("<Hmm, this is a little-endian machine, fixing byte ordering....");
-        reverseACD();			/* Reverse content of the ACD file */
-        if (debugOption||sectionTraceOption||singleStepOption)
-            output("OK.>$n");
-    }
-}
-
-
-/*----------------------------------------------------------------------*/
-static void setupHeader(ACodeHeader tmphdr) {
-    if (!isPreBeta2(tmphdr.version))
-        header = (ACodeHeader *) pointerTo(0);
-    else {
-        header = duplicate(&memory[0], sizeof(ACodeHeader));
-        if (isPreAlpha5(tmphdr.version)) {
-	    header->ifids = 0;
-        }
-        header->prompt = 0;
-    }        
-}
-
-
-/*----------------------------------------------------------------------*/
-static void loadAndCheckMemory(ACodeHeader tmphdr, Aword crc, char err[]) {
-    int i;
-    /* No memory allocated yet? */
-    if (memory == NULL) {
-        memory = allocate(tmphdr.size*sizeof(Aword));
-    }
-
-    memTop = fread(pointerTo(0), sizeof(Aword), tmphdr.size, codfil);
-    if (memTop != tmphdr.size)
-        syserr("Could not read all ACD code.");
-	
-    /* Calculate checksum */
-    for (i = memoryStart(tmphdr.version); i < memTop; i++) {
-        crc += memory[i]&0xff;
-        crc += (memory[i]>>8)&0xff;
-        crc += (memory[i]>>16)&0xff;
-        crc += (memory[i]>>24)&0xff;
-#ifdef CRCLOG
-        printf("%6x\t%6lx\t%6lx\n", i, crc, memory[i]);
-#endif
-    }
-    if (crc != tmphdr.acdcrc) {
-        sprintf(err, "Checksum error in Acode (.a3c) file (0x%lx instead of 0x%lx).",
-                (unsigned long) crc, (unsigned long) tmphdr.acdcrc);
-        if (!ignoreErrorOption)
-            syserr(err);
-        else {
-            output("<WARNING! $$");
-            output(err);
-            output("$$ Ignored, proceed at your own risk.>$n");
-        }
-    }
-}
-
-
 /*----------------------------------------------------------------------*/
 static void load(void)
 {
     ACodeHeader tmphdr;
     Aword crc = 0;
     char err[100];
-	
+
     readTemporaryHeader(&tmphdr);
     checkVersion(&tmphdr);
-	
+
     /* Allocate and load memory */
-	
+
     if (littleEndian())
         reverseHdr(&tmphdr);
-	
+
     if (tmphdr.size <= sizeof(ACodeHeader)/sizeof(Aword))
         syserr("Malformed game file. Too small.");
-	
+
     loadAndCheckMemory(tmphdr, crc, err);
 
     reverseMemory();
@@ -399,17 +418,17 @@ static void checkDebug(void)
 {
     /* Make sure he can't debug if not allowed! */
     if (!header->debug) {
-        if (debugOption|sectionTraceOption|singleStepOption) {
+        if (debugOption|traceSectionOption|traceInstructionOption) {
             printf("<Sorry, '%s' is not compiled for debug! Exiting.>\n", adventureFileName);
             terminate(0);
         }
         para();
         debugOption = FALSE;
-        sectionTraceOption = FALSE;
-        singleStepOption = FALSE;
+        traceSectionOption = FALSE;
+        traceInstructionOption = FALSE;
         tracePushOption = FALSE;
     }
-	
+
     if (debugOption || regressionTestOption) /* If debugging... */
         srand(1);			/* ... use no randomization */
     else
@@ -424,49 +443,49 @@ static void initStaticData(void)
     dictionary = (DictionaryEntry *) pointerTo(header->dictionary);
     /* Find out number of entries in dictionary */
     for (dictionarySize = 0; !isEndOfArray(&dictionary[dictionarySize]); dictionarySize++);
-	
+
     /* Scores */
-	
-	
+
+
     /* All addresses to tables indexed by ids are converted to pointers,
        then adjusted to point to the (imaginary) element before the
        actual table so that [0] does not exist. Instead indices goes
        from 1 and we can use [1]. */
-	
+
     if (header->instanceTableAddress == 0)
         syserr("Instance table pointer == 0");
     instances = (InstanceEntry *) pointerTo(header->instanceTableAddress);
     instances--;			/* Back up one so that first is no. 1 */
-	
-	
+
+
     if (header->classTableAddress == 0)
         syserr("Class table pointer == 0");
     classes = (ClassEntry *) pointerTo(header->classTableAddress);
     classes--;			/* Back up one so that first is no. 1 */
-	
+
     if (header->containerTableAddress != 0) {
         containers = (ContainerEntry *) pointerTo(header->containerTableAddress);
         containers--;
     }
-	
+
     if (header->eventTableAddress != 0) {
         events = (EventEntry *) pointerTo(header->eventTableAddress);
         events--;
     }
-	
+
     /* Scores, if already allocated, copy initial data */
     if (scores == NULL)
         scores = duplicate((Aword *) pointerTo(header->scores), header->scoreCount*sizeof(Aword));
     else
         memcpy(scores, pointerTo(header->scores), header->scoreCount*sizeof(Aword));
-	
+
     if (literals == NULL)
         literals = allocate(sizeof(Aword)*(MAXPARAMS+1));
-	
+
     stxs = (SyntaxEntry *) pointerTo(header->syntaxTableAddress);
     vrbs = (VerbEntry *) pointerTo(header->verbTableAddress);
     msgs = (MessageEntry *) pointerTo(header->messageTableAddress);
-    initRules();
+    initRules(header->ruleTableAddress);
 
     if (header->pack)
         freq = (Aword *) pointerTo(header->freq);
@@ -477,9 +496,9 @@ static void initStaticData(void)
 static void initStrings(void)
 {
     StringInitEntry *init;
-	
+
     for (init = (StringInitEntry *) pointerTo(header->stringInitTable); !isEndOfArray(init); init++)
-        setInstanceAttribute(init->instanceCode, init->attributeCode, (Aptr)getStringFromFile(init->fpos, init->len));
+        setInstanceAttribute(init->instanceCode, init->attributeCode, toAptr(getStringFromFile(init->fpos, init->len)));
 }
 
 /*----------------------------------------------------------------------*/
@@ -487,7 +506,7 @@ static Aint sizeOfAttributeData(void)
 {
     int i;
     int size = 0;
-	
+
     for (i=1; i<=header->instanceMax; i++) {
         AttributeEntry *attribute = pointerTo(instances[i].initialAttributes);
         while (!isEndOfArray(attribute)) {
@@ -496,7 +515,7 @@ static Aint sizeOfAttributeData(void)
         }
         size += 1;			/* For EOF */
     }
-	
+
     if (size != header->attributesAreaSize
         && (sizeof(AttributeHeaderEntry) == sizeof(AttributeEntry)))
         syserr("Attribute area size calculated wrong.");
@@ -510,7 +529,7 @@ static AttributeEntry *initializeAttributes(int awordSize)
     Aword *attributeArea = allocate(awordSize*sizeof(Aword));
     Aword *currentAttributeArea = attributeArea;
     int i;
-	
+
     for (i=1; i<=header->instanceMax; i++) {
         AttributeHeaderEntry *originalAttribute = pointerTo(instances[i].initialAttributes);
         admin[i].attributes = (AttributeEntry *)currentAttributeArea;
@@ -524,7 +543,7 @@ static AttributeEntry *initializeAttributes(int awordSize)
         *((Aword*)currentAttributeArea) = EOF;
         currentAttributeArea += 1;
     }
-	
+
     return (AttributeEntry *)attributeArea;
 }
 
@@ -535,17 +554,17 @@ static AttributeEntry *initializeAttributes(int awordSize)
 static void initDynamicData(void)
 {
     int instanceId;
-	
+
     /* Allocate for administrative table */
     admin = (AdminEntry *)allocate((header->instanceMax+1)*sizeof(AdminEntry));
-	
+
     /* Create game state copy of attributes */
     attributes = initializeAttributes(sizeOfAttributeData());
-	
+
     /* Initialise string & set attributes */
     initStrings();
     initSets((SetInitEntry*)pointerTo(header->setInitTable));
-	
+
     /* Set initial locations */
     for (instanceId = 1; instanceId <= header->instanceMax; instanceId++)
         admin[instanceId].location = instances[instanceId].initialLocation;
@@ -572,7 +591,7 @@ static void runInitialize(Aint theInstance) {
 /*----------------------------------------------------------------------*/
 static void initializeInstances() {
     int instanceId;
-	
+
     /* Set initial locations */
     for (instanceId = 1; instanceId <= header->instanceMax; instanceId++) {
         current.instance = instanceId;
@@ -585,22 +604,25 @@ static void initializeInstances() {
 static void start(void)
 {
     int startloc;
-	
-    current.tick = -1;
-    current.location = startloc = where(HERO, FALSE);
+
+    current.tick = 0;
+    current.location = startloc = where(HERO, TRANSITIVE);
     current.actor = HERO;
     current.score = 0;
-	
+
     initializeInstances();
-	
-    if (sectionTraceOption)
+
+    if (traceSectionOption)
         printf("\n<START:>\n");
     interpret(header->start);
     para();
 
-    if (where(HERO, FALSE) == startloc)
+    if (where(HERO, TRANSITIVE) == startloc) {
+        if (traceSectionOption)
+            printf("<CURRENT LOCATION:>");
         look();
-    evaluateRules();
+    }
+    resetAndEvaluateRules(rules, header->version);
 }
 
 
@@ -608,7 +630,7 @@ static void start(void)
 static void openFiles(void)
 {
     char str[256];
-	
+
     /* Open Acode file */
     strcpy(codfnm, adventureFileName);
     if ((codfil = fopen(codfnm, READ_MODE)) == NULL) {
@@ -617,7 +639,7 @@ static void openFiles(void)
         strcat(str, "'.");
         playererr(str);
     }
-	
+
     /* Open Text file */
     strcpy(txtfnm, adventureFileName);
     if ((textFile = fopen(txtfnm, READ_MODE)) == NULL) {
@@ -626,7 +648,7 @@ static void openFiles(void)
         strcat(str, "'.");
         apperr(str);
     }
-	
+
     /* If logging open log file */
     if (transcriptOption || logOption) {
         startTranscript();
@@ -638,45 +660,29 @@ static void openFiles(void)
 static void init(void)
 {
     int i;
-	
-    if (!regressionTestOption && (debugOption||sectionTraceOption||singleStepOption)) {
-        char str[80];
-        output("<Hi! This is Alan interactive fiction interpreter Arun,");
-        sprintf(str, "version %ld.%ld%s%ld",
-                (long)alan.version.version,
-                (long)alan.version.revision,
-                alan.version.state[0]=='\0'?".":alan.version.state,
-                (long)alan.version.correction);
-        output(str);
-	if (BUILD != 0)
-	    sprintf(str, "- build %d!>$n", BUILD);
-	else
-	    sprintf(str, "!>$n");
-	output(str);
-    }
-	
+
     /* Initialise some status */
     eventQueueTop = 0;			/* No pending events */
     initStaticData();
     initDynamicData();
     initParsing();
     checkDebug();
-	
+
     getPageSize();
-	
+
     /* Find first conjunction and use that for ',' handling */
     for (i = 0; i < dictionarySize; i++)
         if (isConjunction(i)) {
             conjWord = i;
             break;
         }
-	
+
     /* Start the adventure */
     if (debugOption)
         debug(FALSE, 0, 0);
     else
         clear();
-	
+
     start();
 }
 
@@ -685,18 +691,18 @@ static void init(void)
 /*----------------------------------------------------------------------*/
 static bool traceActor(int theActor)
 {
-    if (sectionTraceOption) {
+    if (traceSectionOption) {
         printf("\n<ACTOR ");
         traceSay(theActor);
         printf("[%d]", theActor);
-	if (current.location != 0) {
-	    printf(" (at ");
-	    traceSay(current.location);
-	} else
-	    printf(" (nowhere");
+        if (current.location != 0) {
+            printf(" (at ");
+            traceSay(current.location);
+        } else
+            printf(" (nowhere");
         printf("[%d])", current.location);
     }
-    return sectionTraceOption;
+    return traceSectionOption;
 }
 
 
@@ -704,7 +710,7 @@ static bool traceActor(int theActor)
 static char *scriptName(int theActor, int theScript)
 {
     ScriptEntry *scriptEntry = pointerTo(header->scriptTableAddress);
-	
+
     while (theScript > 1) {
         scriptEntry++;
         theScript--;
@@ -718,15 +724,15 @@ static void moveActor(int theActor)
 {
     ScriptEntry *scr;
     StepEntry *step;
-    Aword previousInstance = current.instance;
-	
+    Aint previousInstance = current.instance;
+
     current.actor = theActor;
     current.instance = theActor;
-    current.location = where(theActor, FALSE);
+    current.location = where(theActor, TRANSITIVE);
     if (theActor == HERO) {
         /* Ask him! */
         if (setjmp(forfeitLabel) == 0) {
-            parse(globalParameters);
+            parse();
             capitalize = TRUE;
             fail = FALSE;			/* fail only aborts one actor */
         }
@@ -767,8 +773,8 @@ static void moveActor(int theActor)
                            (long)admin[theActor].step);
                 interpret(step->stms);
                 step++;
-                /* ... so that we can see if he is USEing another script now */
-                if (admin[theActor].step != 0 && isEndOfArray(step))
+                /* ... so that we can see if he failed or is USEing another script now */
+                if (fail || (admin[theActor].step != 0 && isEndOfArray(step)))
                     /* No more steps in this script, so stop him */
                     admin[theActor].script = 0;
                 fail = FALSE;			/* fail only aborts one actor */
@@ -782,7 +788,7 @@ static void moveActor(int theActor)
             printf(", Idle>\n");
         }
     }
-	
+
     current.instance = previousInstance;
 }
 
@@ -793,32 +799,33 @@ static void moveActor(int theActor)
 void run(void)
 {
     int i;
-    bool playerChangedState;
     static Stack theStack = NULL; /* Needs to survive longjmp() */
-	
+
     openFiles();
     load();			/* Load program */
-	
+
     if (RESTARTED) {
         deleteStack(theStack);
     }
-	
+
     theStack = createStack(STACKSIZE);
     setInterpreterStack(theStack);
-	
+
     initStateStack();
-	
-    if (!ERROR_RETURNED)      /* Can happen in start section to... */
+
+    if (!ERROR_RETURNED)      /* Can happen in start section too... */
         init();               /* Initialise and start the adventure */
-	
+
     while (TRUE) {
         if (debugOption)
             debug(FALSE, 0, 0);
-		
-        runPendingEvents();
 
-        current.tick++;
-		
+        if (stackDepth(theStack) != 0)
+            syserr("Stack is not empty in main loop");
+
+        if (!current.meta)
+            runPendingEvents();
+
         /* Return here if error during execution */
         switch (setjmp(returnLabel)) {
         case NO_JUMP_RETURN:
@@ -833,34 +840,31 @@ void run(void)
         default:
             syserr("Unexpected longjmp() return value");
         }
-		
+
         recursionDepth = 0;
-		
+
         /* Move all characters, hero first */
         rememberGameState();
-		
-        /* TODO: Why 'playerChangedState' since gameStateChanged is sufficient
-         * Actually it isn't since it might have been one of the events or other actors
-         * that changed the state. Why is this important?
-         * Yes, because for UNDO we want to undo the last command the player did that
-         * changed the state, not any of the others. */
-        playerChangedState = FALSE;
+        current.meta = FALSE;
         moveActor(header->theHero);
-        playerChangedState = gameStateChanged;
-		
+
         if (gameStateChanged)
             rememberCommands();
         else
             forgetGameState();
-		
-        evaluateRules();
-        for (i = 1; i <= header->instanceMax; i++)
-            if (i != header->theHero && isActor(i)) {
-                moveActor(i);
-                evaluateRules();
-            }
+
+        if (!current.meta) {
+            current.tick++;
+
+            /* Remove this call? Since Eval is done up there after each event... */
+            resetAndEvaluateRules(rules, header->version);
+
+            /* Then all the other actors... */
+            for (i = 1; i <= header->instanceMax; i++)
+                if (i != header->theHero && isAActor(i)) {
+                    moveActor(i);
+                    resetAndEvaluateRules(rules, header->version);
+                }
+        }
     }
-#ifdef SMARTALLOC
-    sm_dump(1);
-#endif
 }
