@@ -3,11 +3,11 @@
 bool verboseOption = FALSE;
 bool ignoreErrorOption = FALSE;
 bool debugOption = FALSE;
-bool sectionTraceOption = FALSE;
+bool traceSectionOption = FALSE;
 bool tracePushOption = FALSE;
 bool traceStackOption = FALSE;
 bool traceSourceOption = FALSE;
-bool singleStepOption = FALSE;
+bool traceInstructionOption = FALSE;
 bool transcriptOption = FALSE;
 bool logOption = FALSE;
 bool statusLineOption = TRUE;
