@@ -14,10 +14,6 @@
 #include "memory.h"
 
 
-/* MEMORY DEBUGGING? */
-#include "smartall.h"
-
-
 /* PREPROCESSOR */
 #define FORWARD
 #define NEW(type) ((type *)allocate(sizeof(type)))
@@ -25,16 +21,6 @@
 
 /* CONSTANTS */
 
-#ifndef TRUE
-#define TRUE (0==0)
-#endif
-#ifndef FALSE
-#define FALSE (!TRUE)
-#endif
-
-#define LITMIN (header->locmax+1)
-#define LITMAX (header->locmax+1+litCount)
-
 #define HERO (header->theHero)
 #define ENTITY (header->entityClassId)
 #define OBJECT (header->objectClassId)
@@ -43,19 +29,21 @@
 #define ACTOR (header->actorClassId)
 
 #define MAXPARAMS (header->maxParameters)
-#define MAXENTITY (header->instanceMax)
+#define MAXINSTANCE (header->instanceMax)
 
 #define pointerTo(x) ((void *)&memory[x])
 #define addressOf(x) ((((long)x)-((long)memory))/sizeof(Aword))
 #define stringAt(x) ((char *)pointerTo(x))
 
+#define ASIZE(x) (sizeof(x)/sizeof(Aword))
+
 /* TYPES */
 
 #include <stdbool.h>
 
 /* The various tables */
 typedef struct VerbEntry {	/* VERB TABLE */
-  Aword code;			/* Code for the verb */
+  Aint code;			/* Code for the verb */
   Aaddr alts;			/* Address to alternatives */
 } VerbEntry;
 
@@ -65,4 +53,8 @@ typedef struct LimEntry {	/* LIMIT Type */
   Aaddr stms;			/* Statements if fail */
 } LimitEntry;
 
+
+/* Functions: */
+extern Aaddr addressAfterTable(Aaddr adr, int size);
+
 #endif
