@@ -6,16 +6,16 @@
 
 #include "io.h"
 
-#define BLORB_PICT	0x50696374
-#define BLORB_SND	0x536e6420
-#define BLORB_EXEC	0x45786563
+#define BLORB_PICT	0x50696374UL
+#define BLORB_SND	0x536e6420UL
+#define BLORB_EXEC	0x45786563UL
 
 typedef struct zterp_blorb zterp_blorb;
 
 typedef struct
 {
   uint32_t usage;
-  int number;
+  uint32_t number;
   uint32_t type;
   char name[5];
   uint32_t offset;
@@ -24,6 +24,6 @@ typedef struct
 
 zterp_blorb *zterp_blorb_parse(zterp_io *);
 void zterp_blorb_free(zterp_blorb *);
-const zterp_blorb_chunk *zterp_blorb_find(zterp_blorb *, uint32_t, int);
+const zterp_blorb_chunk *zterp_blorb_find(zterp_blorb *, uint32_t, uint32_t);
 
 #endif
