@@ -2,12 +2,15 @@
 #define ZTERP_STACK_H
 
 #include <stdint.h>
+#include <stdbool.h>
 
 #include "io.h"
 
 #define DEFAULT_STACK_SIZE	0x4000
 #define DEFAULT_CALL_DEPTH	0x400
 
+extern bool seen_save_undo;
+
 void init_stack(void);
 
 uint16_t variable(uint16_t);
@@ -20,14 +23,14 @@ uint16_t direct_call(uint16_t);
 #endif
 void do_return(uint16_t);
 
-int save_quetzal(zterp_io *, int);
-int restore_quetzal(zterp_io *, int);
+bool do_save(bool);
+bool do_restore(bool);
 
-int do_save(int);
-int do_restore(int);
-
-int push_save(void);
-int pop_save(void);
+enum save_type { SAVE_GAME, SAVE_USER };
+bool push_save(enum save_type, uint32_t, const char *);
+bool pop_save(enum save_type, long);
+bool drop_save(enum save_type, long);
+void list_saves(enum save_type, void (*)(const char *));
 
 void zpush(void);
 void zpull(void);
