@@ -33,8 +33,18 @@ Opcode* gOpcodeTable;
 // -------------------------------------------------------------
 // Floating point support
 
-#define ENCODE_FLOAT(f) (* (git_uint32*) &f)
-#define DECODE_FLOAT(n) (* (git_float*) &n)
+GIT_INLINE git_uint32 ENCODE_FLOAT(git_float f)
+{
+  git_uint32 n;
+  memcpy(&n, &f, 4);
+  return n;
+}
+
+GIT_INLINE git_float DECODE_FLOAT(git_uint32 n) {
+  git_float f;
+  memcpy(&f, &n, 4);
+  return f;
+}
 
 int floatCompare(git_sint32 L1, git_sint32 L2, git_sint32 L3)
 {
@@ -51,6 +61,9 @@ int floatCompare(git_sint32 L1, git_sint32 L2, git_sint32 L3)
 }
 
 #ifdef USE_OWN_POWF
+float git_powf(float x, float y);
+
+#ifdef GARGLK
 float git_powf(float x, float y)
 {
   if (x == 1.0f)
@@ -63,6 +76,8 @@ float git_powf(float x, float y)
 }
 #endif
 
+#endif
+
 // -------------------------------------------------------------
 // Functions
 
@@ -83,7 +98,6 @@ void startProgram (size_t cacheSize, enum IOMode ioMode)
     git_sint32* top;    // The top of the stack -- that is, the first unusable slot.
 
     git_sint32 args [64]; // Array of arguments. Count is stored in L2.
-    git_uint32 runCounter = 0;
 
     git_uint32 ioRock = 0;
 
@@ -94,8 +108,8 @@ void startProgram (size_t cacheSize, enum IOMode ioMode)
     git_uint32 protectPos = 0;
     git_uint32 protectSize = 0;
     
-    git_uint32 glulxPC = 0;
-    git_uint32 glulxOpcode = 0;
+    git_uint32 maybe_unused glulxPC = 0;
+    git_uint32 maybe_unused glulxOpcode = 0;
 
     acceleration_func accelfunc;
 
@@ -130,12 +144,11 @@ void startProgram (size_t cacheSize, enum IOMode ioMode)
     goto do_enter_function_L1;
 
 #ifdef USE_DIRECT_THREADING
-#define NEXT do { ++runCounter; goto **(pc++); } while(0)
+#define NEXT do { goto **(pc++); } while(0)
 #else
 #define NEXT goto next
 //#define NEXT do { CHECK_USED(0); CHECK_FREE(0); goto next; } while (0)
 next:
-    ++runCounter;
     switch (*pc++)
     {
 #define LABEL(foo) case label_ ## foo: goto do_ ## foo;
@@ -201,15 +214,11 @@ do_S1_addr8:  memWrite8 (READ_PC, S1); NEXT;
 #define UL7 ((git_uint32)L7)
 
 do_recompile:
-    gBlockHeader->runCounter = runCounter;
     pc = compile (READ_PC);
-    runCounter = 0;
 	NEXT;
 	
 do_jump_abs_L7:
-    gBlockHeader->runCounter = runCounter;
     pc = getCode (UL7);
-    runCounter = gBlockHeader->runCounter;
     NEXT;
 
 do_enter_function_L1: // Arg count is in L2.
@@ -725,20 +734,21 @@ do_tailcall:
     resume_number_L7_digit_L6:
     {
         char buffer [16];
+        git_uint32 absn;
         
         // If the IO mode is 'null', do nothing.
         if (ioMode == IO_NULL)
             goto do_pop_call_stub;
 
         // Write the number into the buffer.
-        L1 = (L7 < 0) ? -L7 : L7; // Absolute value of number.
-        L2 = 0;                   // Current buffer position.
+        absn = (L7 < 0) ? -L7 : L7; // Absolute value of number.
+        L2 = 0;                     // Current buffer position.
         do
         {
-            buffer [L2++] = '0' + (L1 % 10);
-            L1 /= 10;
+            buffer [L2++] = '0' + (absn % 10);
+            absn /= 10;
         }
-        while (L1 > 0);
+        while (absn > 0);
 
         if (L7 < 0)
             buffer [L2++] = '-';
@@ -1280,10 +1290,10 @@ do_tailcall:
     
     do_mzero:
         if (L1 > 0) {
-			if (L2 < gRamStart || (L2 + L1) > gEndMem)
-				memWriteError(L2);
-			memset(gRam + L2, 0, L1);
-		}
+          if (L2 < gRamStart || (L2 + L1) > gEndMem)
+            memWriteError(L2);
+          memset(gMem + L2, 0, L1);
+        }
         NEXT;
         
     do_mcopy:
@@ -1292,19 +1302,7 @@ do_tailcall:
                 memReadError(L2);
             if (L3 < gRamStart || (L3 + L1) > gEndMem)
                 memWriteError(L3);
-            // ROM and ROM are stored separately, so this is a bit fiddly...
-            if (L2 > gRamStart) {
-                // Only need to copy from RAM. Might be overlapping, so use memmove.
-                memmove(gRam + L3, gRam + L2, L1);
-            } else if ((L2 + L1) <= gRamStart) {
-                // Only need to copy from ROM. Can't overlap, so memcpy is safe.
-                memcpy(gRam + L3, gRom + L2, L1);
-            } else {
-                // Need to copy from both ROM and RAM.
-                L4 = (L2 + L1) - gRamStart; // Amount of ROM to copy.
-                memcpy(gRam + L3, gRom + L2, L4);
-                memmove(gRam + L3 + L4, gRam + L2 + L4, L1 - L4);
-            }
+            memmove(gMem + L3, gMem + L2, L1);
         }
         NEXT;
         
