@@ -50,7 +50,8 @@
 	#define NOT_CLEVER "It wasn't so clever as to be worth repeating.^"
 	#define NO_MOVES "But you haven't done anything yet!^"
 	#define TYPE_NUMBER "^Type a number between %d and %d: "
-	#define REFERRING_TO "Are you referring to:^"
+	#define BY "By ~"
+	#define REFERRING_TO "~, are you referring to:^"
 	#define WALKTHRU_WORD "walkthru"
 	#define INFO_WORD "info"
 	#define RESTART_WORD "restart"
@@ -123,7 +124,8 @@
 	#define NO_IT_END "~.^"
 	#define BACK_REFERENCE "You must have referred to a noun previously in the same sentence to use the word ~"
 	#define BACK_REFERENCE_END "~.^"
-	#define MUST_SPECIFY "You must specify whether you mean "
+	#define WHEN_YOU_SAY "When you say ~"
+	#define MUST_SPECIFY "~, you must specify whether you mean "
 	#define OR_WORD " or "
 
 	/* SYSTEM MESSAGES */
@@ -147,7 +149,7 @@
 	#define USED_LABEL_OBJ "In line %d, \"%s\" is already used as an object or location label."
 	#define NO_OBJECT_ERR "In line %d, property \"%s\" defined before first object or location."
 	#define BAD_INCLUDE "'#include' directive must be followed by file name enclosed in double quotes."
-	#define BAD_PARAMETER "Unknown container '%s' associated with parameter '%s'."
+	#define BAD_PARAMETER "Unknown or inappropriate type of container '%s' associated with parameter '%s'."
 	#define BAD_VALUE "Value '%s' cannot be stored in container '%s'."
 	#define NO_MEDIA "WebJACL: Media file \"%s\" not found, external media support disabled.\n"
 	#define MEDIA_REGISTERED "WebJACL: Registered %d media.\n"
@@ -162,6 +164,8 @@
 	#define SELF_REFERENCE "In function \"%s\", reference to object \"%s\" whose parent is itself."
 	#define EXECUTING "Executing function \"%s\".\n"
 	#define NO_WHILE "In function \"%s\", 'endwhile' command without matching 'while' command."
+	#define NO_ITERATE "In function \"%s\", 'enditerate' command without matching 'iterate' command."
+	#define NO_UPDATE "In function \"%s\", 'endupdate' command without matching 'update' command."
 	#define NO_REPEAT "In function \"%s\", 'until' command without matching 'repeat' command."
 	#define NO_LOOP "In function \"%s\", 'endloop' command without matching 'loop' command."
 	#define UNDEFINED_FUNCTION "In function \"%s\", attempt to execute undefined function \"%s\"."
@@ -219,6 +223,7 @@
  	#define NOT_CLEVER "Ce n'tait pas si intelligent au point de vouloir rpter cela.^"
  	#define NO_MOVES "Mais vous n'avez rien fait pour le moment !^"
  	#define TYPE_NUMBER "^Entrez un nombre entre %d et %d: "
+	#define BY "By ~"
  	#define REFERRING_TO "Faites-vous rfrence  :^"
  	#define WALKTHRU_WORD "solution"
  	#define INFO_WORD "info"
@@ -292,7 +297,8 @@
  	#define NO_IT_END "~.^"
  	#define BACK_REFERENCE "Vous devez pralablement avoir fait rfrence  un nom reconnu dans la mme phrase pour pouvoir utiliser le mot ~"
  	#define BACK_REFERENCE_END "~.^"
- 	#define MUST_SPECIFY "Vous devez spcifier si vous voulez dire "
+	#define WHEN_YOU_SAY "When you say ~"
+ 	#define MUST_SPECIFY "~, vous devez spcifier si vous voulez dire "
  	#define OR_WORD " ou "
 
 	/* SYSTEM MESSAGES */
@@ -331,6 +337,8 @@
 	#define SELF_REFERENCE "In function \"%s\", reference to object \"%s\" whose parent is itself."
 	#define EXECUTING "Executing function \"%s\".\n"
 	#define NO_WHILE "In function \"%s\", 'endwhile' command without matching 'while' command."
+	#define NO_ITERATE "In function \"%s\", 'enditerate' command without matching 'iterate' command."
+	#define NO_UPDATE "In function \"%s\", 'endupdate' command without matching 'update' command."
 	#define NO_REPEAT "In function \"%s\", 'until' command without matching 'repeat' command."
 	#define NO_LOOP "In function \"%s\", 'endloop' command without matching 'loop' command."
 	#define UNDEFINED_FUNCTION "In function \"%s\", attempt to execute undefined function \"%s\"."
@@ -390,7 +398,8 @@
 	#define NOT_CLEVER "It wasn't so clever as to be worth repeating.^"
 	#define NO_MOVES "But you haven't done anything yet!^"
 	#define TYPE_NUMBER "^Type a number between %d and %d: "
-	#define REFERRING_TO "Are you referring to:^"
+	#define BY "By ~"
+	#define REFERRING_TO "~, are you referring to:^"
 	#define WALKTHRU_WORD "walkthru"
 	#define INFO_WORD "info"
 	#define RESTART_WORD "restart"
@@ -463,7 +472,8 @@
 	#define NO_IT_END "~.^"
 	#define BACK_REFERENCE "You must have referred to a noun previously in the same sentence to use the word ~"
 	#define BACK_REFERENCE_END "~.^"
-	#define MUST_SPECIFY "You must specify whether you mean "
+	#define WHEN_YOU_SAY "When you say ~"
+	#define MUST_SPECIFY "~, you must specify whether you mean "
 	#define OR_WORD " or "
 
 	/* SYSTEM MESSAGES */
@@ -487,7 +497,7 @@
 	#define USED_LABEL_OBJ "In line %d, \"%s\" is already used as an object or location label."
 	#define NO_OBJECT_ERR "In line %d, property \"%s\" defined before first object or location."
 	#define BAD_INCLUDE "'#include' directive must be followed by file name enclosed in double quotes."
-	#define BAD_PARAMETER "Unknown container '%s' associated with parameter '%s'."
+	#define BAD_PARAMETER "Unknown or inappropriate type of container '%s' associated with parameter '%s'."
 	#define BAD_VALUE "Value '%s' cannot be stored in container '%s'."
 	#define NO_MEDIA "WebJACL: Media file \"%s\" not found, external media support disabled.\n"
 	#define MEDIA_REGISTERED "WebJACL: Registered %d media.\n"
@@ -502,6 +512,8 @@
 	#define SELF_REFERENCE "In function \"%s\", reference to object \"%s\" whose parent is itself."
 	#define EXECUTING "Executing function \"%s\".\n"
 	#define NO_WHILE "In function \"%s\", 'endwhile' command without matching 'while' command."
+	#define NO_ITERATE "In function \"%s\", 'enditerate' command without matching 'iterate' command."
+	#define NO_UPDATE "In function \"%s\", 'endupdate' command without matching 'update' command."
 	#define NO_REPEAT "In function \"%s\", 'until' command without matching 'repeat' command."
 	#define NO_LOOP "In function \"%s\", 'endloop' command without matching 'loop' command."
 	#define UNDEFINED_FUNCTION "In function \"%s\", attempt to execute undefined function \"%s\"."
